/*
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * General Public License for more details.
 *
 */

/*
 * Copyright (c) 2000 VERITAS Software Corporation.
 * 
 */
/****************************************************************************
 *  Header: remcom.c,v 1.34 91/03/09 12:29:49 glenne Exp $
 *
 *  Module name: remcom.c $
 *  Revision: 1.34 $
 *  Date: 91/03/09 12:29:49 $
 *  Contributor:     Lake Stevens Instrument Division$
 *
 *  Description:     low level support for gdb debugger. $
 *
 *  Considerations:  only works on target hardware $
 *
 *  Written by:	     Glenn Engel $
 *  Updated by:	     David Grothe <dave@gcom.com>
 *  ModuleState:     Experimental $
 *
 *  NOTES:	     See Below $
 *
 *  Modified for 386 by Jim Kingdon, Cygnus Support.
 *  Compatibility with 2.1.xx kernel by David Grothe <dave@gcom.com>
 *
 *  Changes to allow auto initilization.  All that is needed is that it
 *  be linked with the kernel and a break point (int 3) be executed.
 *  The header file <asm/kgdb.h> defines BREAKPOINT to allow one to do
 *  this. It should also be possible, once the interrupt system is up, to
 *  call putDebugChar("+").  Once this is done, the remote debugger should
 *  get our attention by sending a ^C in a packet. George Anzinger 
 *  <george@mvista.com>
 *  Integrated into 2.2.5 kernel by Tigran Aivazian <tigran@sco.com>
 *  Added thread support, support for multiple processors,
 *	support for ia-32(x86) hardware debugging.
 *	Amit S. Kale ( akale@veritas.com )
 *
 *
 *  To enable debugger support, two things need to happen.  One, a
 *  call to set_debug_traps() is necessary in order to allow any breakpoints
 *  or error conditions to be properly intercepted and reported to gdb.
 *  Two, a breakpoint needs to be generated to begin communication.  This
 *  is most easily accomplished by a call to breakpoint().  Breakpoint()
 *  simulates a breakpoint by executing an int 3.
 *
 *************
 *
 *    The following gdb commands are supported:
 *
 * command          function                               Return value
 *
 *    g             return the value of the CPU registers  hex data or ENN
 *    G             set the value of the CPU registers     OK or ENN
 *
 *    mAA..AA,LLLL  Read LLLL bytes at address AA..AA      hex data or ENN
 *    MAA..AA,LLLL: Write LLLL bytes at address AA.AA      OK or ENN
 *
 *    c             Resume at current address              SNN   ( signal NN)
 *    cAA..AA       Continue at address AA..AA             SNN
 *
 *    s             Step one instruction                   SNN
 *    sAA..AA       Step one instruction from AA..AA       SNN
 *
 *    k             kill
 *
 *    ?             What was the last sigval ?             SNN   (signal NN)
 *
 * All commands and responses are sent with a packet which includes a
 * checksum.  A packet consists of
 *
 * $<packet info>#<checksum>.
 *
 * where
 * <packet info> :: <characters representing the command or response>
 * <checksum>    :: < two hex digits computed as modulo 256 sum of <packetinfo>>
 *
 * When a packet is received, it is first acknowledged with either '+' or '-'.
 * '+' indicates a successful transfer.  '-' indicates a failed transfer.
 *
 * Example:
 *
 * Host:                  Reply:
 * $m0,10#2a               +$00010203040506070809101112131415#42
 *
 ****************************************************************************/
#define KGDB_VERSION "<20020111.1149.09>"
#include <linux/config.h>
#include <linux/types.h>
#include <asm/string.h>			/* for strcpy */
#include <linux/kernel.h>
#include <linux/sched.h>
#include <asm/vm86.h>
#include <asm/system.h>
#include <asm/ptrace.h>			/* for linux pt_regs struct */
#include <asm/kgdb_local.h>
#include <linux/list.h>
#include <asm/atomic.h>
#include <asm/processor.h>
/* RTAI support needs us to really stop/start interrupts */
#define RTAI
#ifdef hard_save_flags
#undef cli()
#undef sti()
#undef save_flags(x) 
#undef restore_flags(x) 
#define sti() hard_sti()
#define cli() hard_cli()
#define save_flags(x) hard_save_flags(x)
#define restore_flags(x) hard_restore_flags(x)
#endif
/************************************************************************
 *
 * external low-level support routines
 */
typedef void (*Function)(void);		  /* pointer to a function */

/* Thread reference */
typedef unsigned char threadref[8];

extern void putDebugChar(int);	 /* write a single character	  */
extern int getDebugChar(void);	 /* read and return a single char */

/************************************************************************/
/* BUFMAX defines the maximum number of characters in inbound/outbound buffers*/
/* at least NUMREGBYTES*2 are needed for register packets */
/* Longer buffer is needed to list all threads */
#define BUFMAX 1024

char * kgdb_version=KGDB_VERSION;

/*  debug >  0 prints ill-formed commands in valid packets & checksum errors */
int	debug_regs = 0;		/* set to non-zero to print registers */

/* filled in by an external module */
char *gdb_module_offsets;

static const char hexchars[]="0123456789abcdef";

/* Number of bytes of registers.  */
#define NUMREGBYTES 64
/*
 * Note that this register image is in a different order than
 * the register image that Linux produces at interrupt time.
 *
 * Linux's register image is defined by struct pt_regs in ptrace.h.
 * Just why GDB uses a different order is a historical mystery.
 */
enum regnames {_EAX,		/* 0 */
	       _ECX,		/* 1 */
	       _EDX,		/* 2 */
	       _EBX,		/* 3 */
	       _ESP,		/* 4 */
	       _EBP,		/* 5 */
	       _ESI,		/* 6 */
	       _EDI,		/* 7 */
	       _PC 		/* 8 also known as eip */,
	       _PS		/* 9 also known as eflags */,
	       _CS,		/* 10 */
	       _SS,		/* 11 */
	       _DS,		/* 12 */
	       _ES,		/* 13 */
	       _FS,		/* 14 */
	       _GS};		/* 15 */



/***************************  ASSEMBLY CODE MACROS *************************/
/*
 * Put the error code here just in case the user cares.	
 * Likewise, the vector number here (since GDB only gets the signal
 * number through the usual means, and that's not very specific). 
 * The called_from is the return address so he can tell how we entered kgdb.
 * This will allow him to seperate out the various possible entries. 
 */
#define REMOTE_DEBUG 0  // set != to turn on printing (also available in info)

#ifdef CONFIG_SMP
#define IF_SMP(x) x
#define MAX_NO_CPUS 8
#define hold_init hold_on_sstep: 1,
#else
#define IF_SMP(x)
#define hold_init
#endif
struct kgdb_info {
        void * called_from;
        long long entry_tsc;
        int errcode;
        int vector;
        int print_debug_info;
#ifdef CONFIG_SMP
        int hold_on_sstep;
        struct {
                volatile struct task_struct *task;
                int hold;
                struct pt_regs *regs;
        } cpus_waiting[MAX_NO_CPUS];
#endif
} kgdb_info = {hold_init print_debug_info: REMOTE_DEBUG};

#define gdb_i386vector  kgdb_info.vector
#define gdb_i386errcode kgdb_info.errcode
#define waiting_cpus    kgdb_info.cpus_waiting 
#define remote_debug    kgdb_info.print_debug_info
#define hold_cpu(cpu)   kgdb_info.cpus_waiting[cpu].hold
/* gdb locks */


#ifdef _raw_read_unlock  // must use a name that is "define"ed, not an inline
//#error "renaming spin"
#undef spin_lock
#undef spin_trylock
#undef spin_unlock
#define spin_lock	 _raw_spin_lock
#define spin_trylock	 _raw_spin_trylock
#define spin_unlock	 _raw_spin_unlock
#else
//#error "not renaming spin"
#endif

#ifdef _SPIN_LOCK_UNLOCKED
#undef SPIN_LOCK_UNLOCKED
#undef spin_lock_init
#undef spin_lock
#undef spin_is_locked
#undef spin_trylock
#undef spin_unlock_wait
#undef spin_unlock

#define SPIN_LOCK_UNLOCKED _SPIN_LOCK_UNLOCKED
#define spin_lock_init	 _spin_lock_init
#define spin_lock	 _spin_lock
#define spin_is_locked	 _spin_is_locked
#define spin_trylock	 _spin_trylock
#define spin_unlock_wait _spin_unlock_wait
#define spin_unlock	 _spin_unlock

#endif

#ifdef CONFIG_SMP
static int in_kgdb_called;
static spinlock_t waitlocks[MAX_NO_CPUS] = 
			     {[0 ... MAX_NO_CPUS-1]= SPIN_LOCK_UNLOCKED};
/*
 * The following array has the thread pointer of each of the "other"
 * cpus.  We make it global so it can be seen by gdb.
 */
volatile int in_kgdb_entry_log[MAX_NO_CPUS];
volatile struct pt_regs *in_kgdb_here_log[MAX_NO_CPUS];
/*
static spinlock_t continuelocks[MAX_NO_CPUS];
*/
spinlock_t kgdb_spinlock = SPIN_LOCK_UNLOCKED;
/* waiters on our spinlock plus us */
static atomic_t spinlock_waiters = ATOMIC_INIT(1); 
static int spinlock_count = 0;
static int spinlock_cpu = 0;
/*
 * Note we use nested spin locks to account for the case where a break
 * point is encountered when calling a function by user direction from
 * kgdb. Also there is the memory exception recursion to account for.
 * Well, yes, but this lets other cpus thru too.  Lets add a
 * cpu id to the lock.
 */
#define KGDB_SPIN_LOCK(x) if( spinlock_count == 0 || \
			      spinlock_cpu != smp_processor_id()){\
				      atomic_inc(&spinlock_waiters); \
				      while (! spin_trylock(x)) {\
					    in_kgdb(&regs);\
				      }\
				      atomic_dec(&spinlock_waiters); \
				      spinlock_count = 1; \
				      spinlock_cpu = smp_processor_id(); \
			  }else{  \
				      spinlock_count++; \
			  }
#define KGDB_SPIN_UNLOCK(x) if( --spinlock_count == 0) spin_unlock(x)
#define NUM_CPUS smp_num_cpus
#else 
unsigned kgdb_spinlock = 0;
#define KGDB_SPIN_LOCK(x) --*x
#define KGDB_SPIN_UNLOCK(x) ++*x
#define NUM_CPUS 1
#endif

int hex(char ch)
{
  if ((ch >= 'a') && (ch <= 'f')) return (ch-'a'+10);
  if ((ch >= '0') && (ch <= '9')) return (ch-'0');
  if ((ch >= 'A') && (ch <= 'F')) return (ch-'A'+10);
  return (-1);
}

IF_SMP(static void to_gdb(const char * mess));

/* scan for the sequence $<data>#<checksum>	*/
void getpacket(char * buffer)
{
  unsigned char checksum;
  unsigned char xmitcsum;
  int  i;
  int  count;
  char ch;

  do {
    /* wait around for the start character, ignore all other characters */
    while ((ch = (getDebugChar() & 0x7f)) != '$');
    checksum = 0;
    xmitcsum = -1;

    count = 0;

    /* now, read until a # or end of buffer is found */
    while (count < BUFMAX) {
      ch = getDebugChar() & 0x7f;
      if (ch == '#') break;
      checksum = checksum + ch;
      buffer[count] = ch;
      count = count + 1;
      }
    buffer[count] = 0;

    if (ch == '#') {
      xmitcsum = hex(getDebugChar() & 0x7f) << 4;
      xmitcsum += hex(getDebugChar() & 0x7f);
      if ((remote_debug ) && (checksum != xmitcsum)) {
	printk ("bad checksum.	My count = 0x%x, sent=0x%x. buf=%s\n",
		 checksum,xmitcsum,buffer);
      }

      if (checksum != xmitcsum) putDebugChar('-');  /* failed checksum */
      else {
	 putDebugChar('+');  /* successful transfer */
	 /* if a sequence char is present, reply the sequence ID */
	 if (buffer[2] == ':') {
	    putDebugChar( buffer[0] );
	    putDebugChar( buffer[1] );
	    /* remove sequence chars from buffer */
	    count = strlen(buffer);
	    for (i=3; i <= count; i++) buffer[i-3] = buffer[i];
	 }
      }
    }
  } while (checksum != xmitcsum);

  if (remote_debug)
    printk("R:%s\n", buffer) ;
}

/* send the packet in buffer.  */


void putpacket(char * buffer)
{
  unsigned char checksum;
  int  count;
  char ch;

  /*  $<packet info>#<checksum>. */
  do {
  if (remote_debug)
    printk("T:%s\n", buffer) ;
  putDebugChar('$');
  checksum = 0;
  count	   = 0;

  while ((ch=buffer[count])) {
    putDebugChar(ch);
    checksum += ch;
    count += 1;
  }

  putDebugChar('#');
  putDebugChar(hexchars[checksum >> 4]);
  putDebugChar(hexchars[checksum % 16]);

  } while ((getDebugChar() & 0x7f) != '+');

}

static char  remcomInBuffer[BUFMAX];
static char  remcomOutBuffer[BUFMAX];
static short error;


void debug_error( char * format, char * parm)
{
  if (remote_debug) printk (format,parm);
}

static void print_regs(struct pt_regs *regs)
{
    printk("EAX=%08lx ", regs->eax);
    printk("EBX=%08lx ", regs->ebx);
    printk("ECX=%08lx ", regs->ecx);
    printk("EDX=%08lx ", regs->edx);
    printk("\n");
    printk("ESI=%08lx ", regs->esi);
    printk("EDI=%08lx ", regs->edi);
    printk("EBP=%08lx ", regs->ebp);
    printk("ESP=%08lx ", (long) &regs->esp);
    printk("\n");
    printk(" DS=%08x ", regs->xds);
    printk(" ES=%08x ", regs->xes);
    printk(" SS=%08x ", __KERNEL_DS);
    printk(" FL=%08lx ", regs->eflags);
    printk("\n");
    printk(" CS=%08x ", regs->xcs);
    printk(" IP=%08lx ", regs->eip);
#if 0
    printk(" FS=%08x ", regs->fs);
    printk(" GS=%08x ", regs->gs);
#endif
    printk("\n");


} /* print_regs */

static void regs_to_gdb_regs(int *gdb_regs, struct pt_regs *regs)
{
    gdb_regs[_EAX] =  regs->eax;
    gdb_regs[_EBX] =  regs->ebx;
    gdb_regs[_ECX] =  regs->ecx;
    gdb_regs[_EDX] =  regs->edx;
    gdb_regs[_ESI] =  regs->esi;
    gdb_regs[_EDI] =  regs->edi;
    gdb_regs[_EBP] =  regs->ebp;
    gdb_regs[ _DS] =  regs->xds;
    gdb_regs[ _ES] =  regs->xes;
    gdb_regs[ _PS] =  regs->eflags;
    gdb_regs[ _CS] =  regs->xcs;
    gdb_regs[ _PC] =  regs->eip;
    /* Note, as we are a debugging the kernel, we will always 
     * trap in kernel code, this means no priviledge change,
     * and so the pt_regs structure is not completely valid.  In a non
     * privilege change trap, only EFLAGS, CS and EIP are put on the stack,
     * SS and ESP are not stacked, this means that the last 2 elements of
     * pt_regs is not valid (they would normally refer to the user stack)
     * also, using regs+1 is no good because you end up will a value that is 
     * 2 longs (8) too high.  This used to cause stepping over functions
     * to fail, so my fix is to use the address of regs->esp, which 
     * should point at the end of the stack frame.  Note I have ignored
     * completely exceptions that cause an error code to be stacked, such
     * as double fault.	 Stuart Hughes, Zentropix.
     * original code: gdb_regs[_ESP] =	(int) (regs + 1) ; 
     */
    gdb_regs[_ESP] =  (int) (&regs->esp) ;
    gdb_regs[ _SS] =  __KERNEL_DS;
    gdb_regs[ _FS] =  0xFFFF;
    gdb_regs[ _GS] =  0xFFFF;
} /* regs_to_gdb_regs */

static void gdb_regs_to_regs(int *gdb_regs, struct pt_regs *regs)
{
    regs->eax	=     gdb_regs[_EAX] ;
    regs->ebx	=     gdb_regs[_EBX] ;
    regs->ecx	=     gdb_regs[_ECX] ;
    regs->edx	=     gdb_regs[_EDX] ;
    regs->esi	=     gdb_regs[_ESI] ;
    regs->edi	=     gdb_regs[_EDI] ;
    regs->ebp	=     gdb_regs[_EBP] ;
    regs->xds	=     gdb_regs[ _DS] ;
    regs->xes	=     gdb_regs[ _ES] ;
    regs->eflags=     gdb_regs[ _PS] ;
    regs->xcs	=     gdb_regs[ _CS] ;
    regs->eip	=     gdb_regs[ _PC] ;
#if 0					/* can't change these */
    regs->esp	=     gdb_regs[_ESP] ;
    regs->xss	=     gdb_regs[ _SS] ;
    regs->fs	=     gdb_regs[ _FS] ;
    regs->gs	=     gdb_regs[ _GS] ;
#endif

} /* gdb_regs_to_regs */

/* Indicate to caller of mem2hex or hex2mem that there has been an
   error.  */
static volatile int mem_err = 0;
static volatile int mem_err_expected = 0;
static volatile int mem_err_cnt = 0;
static		int garbage_loc = -1 ;

int
get_char (char *addr)
{
  return *addr;
}

void
set_char ( char *addr, int val)
{
  *addr = val;
}

/* convert the memory pointed to by mem into hex, placing result in buf */
/* return a pointer to the last char put in buf (null) */
/* If MAY_FAULT is non-zero, then we should set mem_err in response to
   a fault; if zero treat a fault like any other fault in the stub.  */
char* mem2hex( char* mem,
	       char* buf,
	       int   count,
	       int may_fault)
{
      int i;
      unsigned char ch;

      if (may_fault)
      {
	  mem_err_expected = 1 ;
	  mem_err = 0 ;
      }
      for (i=0;i<count;i++) {
	  /* printk("%lx = ", mem) ; */

	  ch = get_char (mem++);

	  /* printk("%02x\n", ch & 0xFF) ; */
	  if (may_fault && mem_err)
	  {
	    if (remote_debug)
		printk("Mem fault fetching from addr %lx\n", (long)(mem-1));
	    *buf = 0 ;				/* truncate buffer */
	    return (buf);
	  }
	  *buf++ = hexchars[ch >> 4];
	  *buf++ = hexchars[ch % 16];
      }
      *buf = 0;
      if (may_fault)
	  mem_err_expected = 0 ;
      return(buf);
}

/* convert the hex array pointed to by buf into binary to be placed in mem */
/* return a pointer to the character AFTER the last byte written */
char* hex2mem( char* buf,
	       char* mem,
	       int   count,
	       int may_fault)
{
      int i;
      unsigned char ch;

      if (may_fault)
      {
	  mem_err_expected = 1 ;
	  mem_err = 0 ;
      }
      for (i=0;i<count;i++) {
	  ch = hex(*buf++) << 4;
	  ch = ch + hex(*buf++);
	  set_char (mem++, ch);

	  if (may_fault && mem_err)
	  {
	    if (remote_debug)
		printk("Mem fault storing to addr %lx\n", (long)(mem-1));
	    return (mem);
	  }
      }
      if (may_fault)
	  mem_err_expected = 0 ;
      return(mem);
}


/**********************************************/
/* WHILE WE FIND NICE HEX CHARS, BUILD AN INT */
/* RETURN NUMBER OF CHARS PROCESSED	      */
/**********************************************/
int hexToInt(char **ptr, int *intValue)
{
    int numChars = 0;
    int hexValue;

    *intValue = 0;

    while (**ptr)
    {
	hexValue = hex(**ptr);
	if (hexValue >=0)
	{
	    *intValue = (*intValue <<4) | hexValue;
	    numChars ++;
	}
	else
	    break;

	(*ptr)++;
    }

    return (numChars);
}
#define stubhex(h) hex(h)

static int
stub_unpack_int (
     char *buff,
     int fieldlength)
{
  int nibble;
  int retval = 0;

  while (fieldlength)
    {
      nibble = stubhex (*buff++);
      retval |= nibble;
      fieldlength--;
      if (fieldlength)
	retval = retval << 4;
    }
  return retval;
}


static char *
pack_hex_byte (
     char *pkt,
     int byte)
{
  *pkt++ = hexchars[(byte >> 4) & 0xf];
  *pkt++ = hexchars[(byte & 0xf)];
  return pkt;
}

#define BUF_THREAD_ID_SIZE 16

static char *
pack_threadid (
     char *pkt,
     threadref *id)
{
  char *limit;
  unsigned char *altid;

  altid = (unsigned char *) id;
  limit = pkt + BUF_THREAD_ID_SIZE;
  while (pkt < limit)
    pkt = pack_hex_byte (pkt, *altid++);
  return pkt;
}

static char *
unpack_byte (
     char *buf,
     int *value)
{
  *value = stub_unpack_int (buf, 2);
  return buf + 2;
}

static char *
unpack_threadid (
     char *inbuf,
     threadref *id)
{
  char *altref;
  char *limit = inbuf + BUF_THREAD_ID_SIZE;
  int x, y;

  altref = (char *) id;

  while (inbuf < limit)
    {
      x = stubhex (*inbuf++);
      y = stubhex (*inbuf++);
      *altref++ = (x << 4) | y;
    }
  return inbuf;
}

void
int_to_threadref (
     threadref *id,
     int value)
{
  unsigned char *scan;

  scan = (unsigned char *) id;
  {
    int i = 4;
    while (i--)
      *scan++ = 0;
  }
  *scan++ = (value >> 24) & 0xff;
  *scan++ = (value >> 16) & 0xff;
  *scan++ = (value >> 8) & 0xff;
  *scan++ = (value & 0xff);
}

static int
threadref_to_int (
     threadref *ref)
{
  int i, value = 0;
  unsigned char *scan;

  scan = (char *) ref;
  scan += 4;
  i = 4;
  while (i-- > 0)
    value = (value << 8) | ((*scan++) & 0xff);
  return value;
}


extern struct task_struct *kgdb_get_idle(int cpu);
#define idle_task(cpu) kgdb_get_idle(cpu)

struct task_struct *
getthread (
	int pid)
{
	struct task_struct *thread;
	if ( pid >= PID_MAX && pid <= (PID_MAX + NUM_CPUS)){
		return idle_task(pid - PID_MAX);
	}else{
		/*
		 * find_task_by_pid is relatively safe all the time
		 * Other pid functions require lock downs which imply
		 * that we may be interrupting them (as we get here
		 * in the middle of most any lock down)
		 */
		thread = find_task_by_pid(pid);
		if (thread) {
			return thread;
		}
	}
	return NULL;
}

struct hw_breakpoint {
	unsigned enabled;
	unsigned type;
	unsigned len;
	unsigned addr;
} breakinfo[4] = { { enabled:0 }, { enabled:0 }, { enabled:0 }, { enabled:0 }};
unsigned hw_breakpoint_status;
void correct_hw_break( void )
{
	int breakno;
	int correctit;
	int breakbit;
	unsigned dr7;

	asm volatile ( 
		"movl %%db7, %0\n"
		: "=r" (dr7)
		: );
	do
	{
		unsigned addr0, addr1, addr2, addr3;
		asm volatile (
			"movl %%db0, %0\n"
			"movl %%db1, %1\n"
			"movl %%db2, %2\n"
			"movl %%db3, %3\n"
			: "=r" (addr0), "=r" (addr1), "=r" (addr2),
			"=r" (addr3) : );
	} while (0);
	correctit = 0;
	for (breakno = 0; breakno < 3; breakno++) {
		breakbit = 2 << (breakno << 1);
		if (!(dr7 & breakbit) && breakinfo[breakno].enabled) {
			correctit = 1;
			dr7 |= breakbit;
			dr7 &= ~(0xf0000 << (breakno << 2));
			dr7 |= (((breakinfo[breakno].len << 2) |
			       breakinfo[breakno].type) << 16) <<
			       (breakno << 2);
			switch (breakno) {
			case 0:
				asm volatile ("movl %0, %%dr0\n"
					: 
					: "r" (breakinfo[breakno].addr) );
				break;

			case 1:
				asm volatile ("movl %0, %%dr1\n"
					: 
					: "r" (breakinfo[breakno].addr) );
				break;

			case 2:
				asm volatile ("movl %0, %%dr2\n"
					: 
					: "r" (breakinfo[breakno].addr) );
				break;

			case 3:
				asm volatile ("movl %0, %%dr3\n"
					: 
					: "r" (breakinfo[breakno].addr) );
				break;
			}
		} else if ((dr7 & breakbit) && !breakinfo[breakno].enabled){
			correctit = 1;
			dr7 &= ~breakbit;
			dr7 &= ~(0xf0000 << (breakno << 2));
		}
	}
	if (correctit) {
		asm volatile ( "movl %0, %%db7\n"
			       :
			       : "r" (dr7));
	}
}

int remove_hw_break(
	unsigned breakno)
{
	if (!breakinfo[breakno].enabled) {
		return -1;
	}
	breakinfo[breakno].enabled = 0;
	return 0;
}

int set_hw_break(
	unsigned breakno,
	unsigned type,
	unsigned len,
	unsigned addr)
{
	if (breakinfo[breakno].enabled) {
		return -1;
	}
	breakinfo[breakno].enabled = 1;
	breakinfo[breakno].type = type;
	breakinfo[breakno].len = len;
	breakinfo[breakno].addr = addr;
	return 0;
}

#ifdef CONFIG_SMP
static int in_kgdb_console=0;

int in_kgdb(struct pt_regs *regs)
{
	unsigned flags;
	int cpu = smp_processor_id();
	in_kgdb_called = 1;
        if (in_kgdb_console) return 1;
	if (!spin_is_locked(&kgdb_spinlock)){
                if ( in_kgdb_here_log[cpu] ){ // we are holding this cpu
                        return 1;
                }
		return 0;
	}

	/* As I see it the only reason not to let all cpus spin on
	 * the same spin_lock is to allow selected ones to proceed.
	 * This would be a good thing, so we leave it this way.
	 * Maybe someday....

	 * in_kgdb() is called from an NMI so we don't pretend
	 * to have any resources, like printk() for example.  
	 */

	local_irq_save(flags); /* only local here, to avoid hanging */
	/*
	 * log arival of this cpu
	 * The NMI keeps on ticking.  Protect against recurring more
	 * than once, and ignor the cpu that has the kgdb lock
	 */
	in_kgdb_entry_log[cpu]++;
	in_kgdb_here_log[cpu] = regs;
	if ( cpu == spinlock_cpu || waiting_cpus[cpu].task ){
		 goto exit_in_kgdb;
	}
	/*
	 * For protection of the initilization of the spin locks by kgdb
	 * it locks the kgdb spinlock before it gets the wait locks set
	 * up.	We wait here for the wait lock to be taken.  If the
	 * kgdb lock goes away first??	Well, it could be a slow exit
	 * sequence where the wait lock is removed prior to the kgdb lock
	 * so if kgdb gets unlocked, we just exit.
	 */
	while (spin_is_locked(&kgdb_spinlock) && 
	       ! spin_is_locked(waitlocks + cpu));
	if ( !spin_is_locked(&kgdb_spinlock)){
		goto exit_in_kgdb;
	}
	waiting_cpus[cpu].task = current;
	waiting_cpus[cpu].regs = regs;
	/*
	 * help out the thread code.  This will tell'em where we are... 
	 */
	current->thread.esp = (int) &regs->esp;
	current->thread.eip = regs->eip;
	spin_unlock_wait(waitlocks + cpu);
	/*
	 * log departure of this cpu
	 */
	waiting_cpus[cpu].task = 0;
	waiting_cpus[cpu].regs = 0;
	correct_hw_break();
 exit_in_kgdb:
	in_kgdb_here_log[cpu] = 0;
	local_irq_restore(flags);
	return 1;
	/*
	  spin_unlock(continuelocks + smp_processor_id());
	*/
}
#else
int in_kgdb(struct pt_regs *regs)
{
	return (kgdb_spinlock);
}
#endif

void
printexceptioninfo(
	int exceptionNo,
	int errorcode,
	char *buffer)
{
	unsigned dr6;
	int i;
	switch (exceptionNo) {
	case 1:			/* debug exception */
		break;
	case 3:			/* breakpoint */
		sprintf(buffer, "Software breakpoint");
		return;
	default:
		sprintf(buffer, "Details not available");
		return;
	}
	asm volatile ("movl %%db6, %0\n"
		      : "=r" (dr6)
		      : );
	if (dr6 & 0x4000) {
		sprintf(buffer, "Single step");
		return;
	}
	for (i = 0; i < 4; ++i) {
		if (dr6 & (1 << i)) {
			sprintf(buffer, "Hardware breakpoint %d", i);
			return;
		}
	}
	sprintf(buffer, "Unknown trap");
	return;
}

/*
 * This function does all command procesing for interfacing to gdb.
 *
 * NOTE:  The INT nn instruction leaves the state of the interrupt
 *	  enable flag UNCHANGED.  That means that when this routine
 *	  is entered via a breakpoint (INT 3) instruction from code
 *	  that has interrupts enabled, then interrupts will STILL BE
 *	  enabled when this routine is entered.	 The first thing that
 *	  we do here is disable interrupts so as to prevent recursive
 *	  entries and bothersome serial interrupts while we are
 *	  trying to run the serial port in polled mode.
 *
 * For kernel version 2.1.xx the cli() actually gets a spin lock so
 * it is always necessary to do a restore_flags before returning
 * so as to let go of that lock.
 */
int handle_exception(int exceptionVector,
		     int signo,
		     int err_code,
		     struct pt_regs *linux_regs)
{
	struct task_struct	*usethread = NULL;
	struct task_struct	*thread = NULL;
	int			addr, length;
	int			breakno, breaktype;
	char			*ptr;
	int			newPC;
	threadref		thref;
	int			threadid;
	int			maxthreads;
	int			nothreads;
	unsigned long		flags;
	int			gdb_regs[NUMREGBYTES/4];
	int			dr6;
        IF_SMP(int              entry_state = 0;) /* 0, ok, 1, no nmi, 2 sync failed */
#define NO_NMI 1
#define NO_SYNC 2
#define	regs	(*linux_regs)
#define NUMREGS NUMREGBYTES/4
	/*
	 * If the entry is not from the kernel then return to the Linux
	 * trap handler and let it process the interrupt normally.
	 */
	if ((linux_regs->eflags & VM_MASK) || (3 & linux_regs->xcs)) {
		printk("ignoring non-kernel exception\n");
		print_regs(&regs);
		return(0);
	}

	save_flags(flags);
	cli();		/* 2.1 kernel must have matching restore_flags */

	/* Get kgdb spinlock */

	KGDB_SPIN_LOCK(&kgdb_spinlock);
        rdtscll(kgdb_info.entry_tsc);
	/*
	 * We depend on this spinlock and the NMI watch dog to control the 
	 * other cpus.	They will arrive at "in_kgdb()" as a result of the
	 * NMI and will wait there for the following spin locks to be 
	 * released.
	 */
#ifdef CONFIG_SMP

	if (smp_num_cpus > MAX_NO_CPUS) {
		printk("kgdb : too many cpus, change MAX_NO_CPUS"
                       " in kgdb_stub and make new kernel.\n");
		goto exit_just_unlock;
	}
	if ( spinlock_count == 1){
		int time,end_time,dum;
		int i;
		int cpu_logged_in[MAX_NO_CPUS] = { [0 ... MAX_NO_CPUS -1]= (0)};
		if (remote_debug){
			printk("kgdb : cpu %d entry, syncing others\n",
			       smp_processor_id());
		}
		for (i = 0; i < smp_num_cpus; i++) {
                        /*
                         * Use trylock as we may already hold the lock if
                         * we are holding the cpu.  Net result is all
                         * locked.
                         */
			spin_trylock(&waitlocks[i]);
		}
                for (i = 0; i < MAX_NO_CPUS; i++) cpu_logged_in[i] = 0;
		/*
		 * Wait for their arrival.  We know the watch dog is active if 
		 * in_kgdb() has ever been called, as it is always called on a 
		 * watchdog tick.
		 */
		rdtsc(dum,time);
		end_time = time + 2; /* Note: we use the High order bits! */
		i=1;
		if ( in_kgdb_called && (smp_num_cpus > 1)){
			int me_in_kgdb = in_kgdb_entry_log[smp_processor_id()];
			while ( i < smp_num_cpus && time != end_time){
				int j;
				for (j = 0; j < MAX_NO_CPUS; j++) {
					if (waiting_cpus[j].task && ! cpu_logged_in[j]){
						i++;
						cpu_logged_in[j] = 1;
						if (remote_debug){
							printk("kgdb : cpu %d arrived at kgdb\n",j);
						}
						break;
					}
					if (!waiting_cpus[j].task && in_kgdb_here_log[j]){
						int wait=100000;
						while(wait--);
						if (!waiting_cpus[j].task && 
						    in_kgdb_here_log[j]){
							printk("kgdb : cpu %d stall"
							       " in in_kgdb\n",j);
							i++;
							cpu_logged_in[j] = 1;
							waiting_cpus[j].task = 
								(struct task_struct *)1;
						}
					}
				}

                                if (in_kgdb_entry_log[smp_processor_id()] > 
				    (me_in_kgdb + 10)){
					break;
				}

				rdtsc(dum,time);
			}
			if (i < smp_num_cpus){
				printk("kgdb : time out, proceeding without sync\n");
#if 0
                                printk("kgdb : Waiting_cpus: 0 = %d, 1 = %d\n",
                                       waiting_cpus[0].task != 0,
                                       waiting_cpus[1].task != 0);
                                printk("kgdb : Cpu_logged in: 0 = %d, 1 = %d\n",
                                       cpu_logged_in[0],
                                       cpu_logged_in[1]);
                                printk("kgdb : in_kgdb_here_log in: 0 = %d, 1 = %d\n",
                                       in_kgdb_here_log[0] != 0,
                                       in_kgdb_here_log[1] != 0);
#endif
                                entry_state = NO_SYNC;
			}else{
#if 0
                                int ent = in_kgdb_entry_log[smp_processor_id()]-
                                        me_in_kgdb;
                                printk("kgdb : sync after %d entries\n",ent);
#endif
                        }
		}else{
			if (remote_debug){
				printk("kgdb : %d cpus, but watchdog not active\n"
				       "proceeding without locking down other cpus\n",
				       smp_num_cpus );
                                entry_state = NO_NMI;
			}
		}
	}
#endif
	  
	if (remote_debug)
		{
			unsigned long	*lp = (unsigned long *) &linux_regs ;

			printk("handle_exception(exceptionVector=%d, "
			       "signo=%d, err_code=%d, linux_regs=%p)\n",
			       exceptionVector, signo, err_code, linux_regs) ;
			if (debug_regs)
				{
					print_regs(&regs) ;
					printk("Stk: %8lx %8lx %8lx %8lx  %8lx %8lx %8lx %8lx\n",
					       lp[0],lp[1],lp[2],lp[3],lp[4],lp[5],lp[6],lp[7]);
					printk("     %8lx %8lx %8lx %8lx  %8lx %8lx %8lx %8lx\n",
					       lp[8],lp[9],lp[10],lp[11],lp[12],lp[13],lp[14],lp[15]);
					printk("     %8lx %8lx %8lx %8lx  %8lx %8lx %8lx %8lx\n",
					       lp[16],lp[17],lp[18],lp[19],lp[20],lp[21],lp[22],lp[23]);
					printk("     %8lx %8lx %8lx %8lx  %8lx %8lx %8lx %8lx\n",
					       lp[24],lp[25],lp[26],lp[27],lp[28],lp[29],lp[30],lp[31]);
				}
		}

	/* Disable hardware debugging while we are in kgdb */
	/* Get the debug register status register */
	__asm__("movl %0,%%db7"
		: /* no output */
		: "r" (0));
	asm volatile ("movl %%db6, %0\n"
		      : "=r" (hw_breakpoint_status)
		      : );

	switch (exceptionVector)
		{
		case 0:			/* divide error */
		case 1:			/* debug exception */
		case 2:			/* NMI */
		case 3:			/* breakpoint */
		case 4:			/* overflow */
		case 5:			/* bounds check */
		case 6:			/* invalid opcode */
		case 7:			/* device not available */
		case 8:			/* double fault (errcode) */
		case 10:			/* invalid TSS (errcode) */
		case 12:			/* stack fault (errcode) */
		case 16:			/* floating point error */
		case 17:			/* alignment check (errcode) */
		default:			/* any undocumented */
			break ;
		case 11:			/* segment not present (errcode) */
		case 13:			/* general protection (errcode) */
		case 14:			/* page fault (special errcode) */
		case 19:			/* cache flush denied */
			if (mem_err_expected)
				{
					/*
					 * This fault occured because of the get_char or set_char
					 * routines.  These two routines use either eax of edx to
					 * indirectly reference the location in memory that they
					 * are working with.  For a page fault, when we return
					 * the instruction will be retried, so we have to make
					 * sure that these registers point to valid memory.
					 */
					mem_err = 1 ;		/* set mem error flag */
					mem_err_expected = 0 ;
					mem_err_cnt++ ;	/* helps in debugging */
					regs.eax = (long) &garbage_loc ;	/* make valid address */
					regs.edx = (long) &garbage_loc ;	/* make valid address */
					if (remote_debug)
						printk("Return after memory error: mem_err_cnt=%d\n",
						       mem_err_cnt);
					if (debug_regs)
						print_regs(&regs) ;
					goto exit_kgdb;
				}
			break ;
		}
	if (remote_debug)
		printk("kgdb : entered kgdb on cpu %d\n", smp_processor_id());

	gdb_i386vector	= exceptionVector;
	gdb_i386errcode = err_code ;
        kgdb_info.called_from = __builtin_return_address (0);
#ifdef CONFIG_SMP
        /*
         * OK, we can now communicate, lets tell gdb about the sync.
         * but only if we had a problem.
         */
        switch (entry_state ) {
        case NO_NMI:
                to_gdb("NMI not active, other cpus not stopped\n");
                break;
        case NO_SYNC:
                to_gdb("Some cpus not stopped, see 'kgdb_info' for details\n");
        default:
        }

#endif

IF_SMP(once_again:)
	/* reply to host that an exception has occurred */
	remcomOutBuffer[0] = 'S';
	remcomOutBuffer[1] =  hexchars[signo >> 4];
	remcomOutBuffer[2] =  hexchars[signo % 16];
	remcomOutBuffer[3] = 0;
 
	putpacket(remcomOutBuffer);

	while (1==1) {
		error = 0;
		remcomOutBuffer[0] = 0;
		getpacket(remcomInBuffer);
		switch (remcomInBuffer[0]) {
		case '?' : 
			remcomOutBuffer[0] = 'S';
			remcomOutBuffer[1] =  hexchars[signo >> 4];
			remcomOutBuffer[2] =  hexchars[signo % 16];
			remcomOutBuffer[3] = 0;
			break;
		case 'd' : remote_debug = !(remote_debug);  /* toggle debug flag */
			printk("Remote debug %s\n", remote_debug ? "on" : "off");
			break;
		case 'g' : /* return the value of the CPU registers */
			if (!usethread || usethread == current) {
				regs_to_gdb_regs(gdb_regs, &regs) ;
			} else {
				memset(gdb_regs, 0, sizeof (gdb_regs));
				gdb_regs[_EBP] =
					gdb_regs[_ESP] = usethread->thread.esp;
				gdb_regs[_PC] = usethread->thread.eip;
			}
			mem2hex((char*) gdb_regs, remcomOutBuffer, NUMREGBYTES, 0);
			break;
		case 'G' : /* set the value of the CPU registers - return OK */
			hex2mem(&remcomInBuffer[1], (char*) gdb_regs, NUMREGBYTES, 0);
			if (!usethread || usethread == current) {
				gdb_regs_to_regs(gdb_regs, &regs) ;
				strcpy(remcomOutBuffer,"OK");
			} else {
				strcpy(remcomOutBuffer,"E00");
			}
			break;

		case 'P' : /* set the value of a single CPU register - return OK */
			{
				int regno;
			
				ptr = &remcomInBuffer[1];
				regs_to_gdb_regs(gdb_regs, &regs) ;
				if ((!usethread || usethread == current) &&
				    hexToInt (&ptr, &regno) && 
				    *ptr++ == '='  &&
				    (regno >= 0 && regno < NUMREGS)) {
					hex2mem (ptr, (char *)&gdb_regs[regno], 4, 0);
					gdb_regs_to_regs(gdb_regs, &regs) ;
					strcpy(remcomOutBuffer,"OK");
					break;
				}
				strcpy (remcomOutBuffer, "E01");
				break;
			}

		/* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */
		case 'm' :
			/* TRY TO READ %x,%x.  IF SUCCEED, SET PTR = 0 */
			ptr = &remcomInBuffer[1];
			if (hexToInt(&ptr,&addr))
				if (*(ptr++) == ',')
					if (hexToInt(&ptr,&length))
						{
							ptr = 0;
							mem2hex((char*) addr, remcomOutBuffer, length, 1);
							if (mem_err) {
								strcpy (remcomOutBuffer, "E03");
								debug_error ("memory fault\n", NULL);
							}
						}

			if (ptr)
				{
					strcpy(remcomOutBuffer,"E01");
					debug_error("malformed read memory command: %s\n",remcomInBuffer);
				}
			break;

			/* MAA..AA,LLLL: Write LLLL bytes at address AA.AA return OK */
		case 'M' :
			/* TRY TO READ '%x,%x:'.  IF SUCCEED, SET PTR = 0 */
			ptr = &remcomInBuffer[1];
			if (hexToInt(&ptr,&addr))
				if (*(ptr++) == ',')
					if (hexToInt(&ptr,&length))
						if (*(ptr++) == ':')
							{
								hex2mem(ptr, (char*) addr, length, 1);

								if (mem_err) {
									strcpy (remcomOutBuffer, "E03");
									debug_error ("memory fault\n", NULL);
								} else {
									strcpy(remcomOutBuffer,"OK");
								}

								ptr = 0;
							}
			if (ptr)
				{
					strcpy(remcomOutBuffer,"E02");
					debug_error("malformed write memory command: %s\n",remcomInBuffer);
				}
			break;

			/* cAA..AA    Continue at address AA..AA(optional) */
			/* sAA..AA   Step one instruction from AA..AA(optional) */
			/* D	    detach, reply OK and then continue */
		case 'c' :
		case 's' :
		case 'D' :
     
			/* try to read optional parameter, pc unchanged if no parm */
			ptr = &remcomInBuffer[1];
			if (hexToInt(&ptr,&addr))
				{
					if (remote_debug)
						printk("Changing EIP to 0x%x\n", addr) ;

					regs.eip = addr;
				}

			newPC = regs.eip ;

			/* clear the trace bit */
			regs.eflags &= 0xfffffeff;

			/* set the trace bit if we're stepping */
			if (remcomInBuffer[0] == 's') regs.eflags |= 0x100;

			/* detach is a friendly version of continue. Note that
			   debugging is still enabled (e.g hit control C)
			   until the process that issued an ioctl TIOCGDB
			   terminates
			*/
			if (remcomInBuffer[0] == 'D') {
				strcpy(remcomOutBuffer,"OK");
				putpacket(remcomOutBuffer);
			}

			if (remote_debug)
				{
					printk("Resuming execution\n") ;
					print_regs(&regs) ;
				}
			asm volatile ("movl %%db6, %0\n"
				      : "=r" (dr6)
				      : );
			if (!(dr6 & 0x4000)) {
				for (breakno = 0; breakno < 4; ++breakno) {
					if (dr6 & (1 << breakno)) {
						if (breakinfo[breakno].type == 0) {
							/* Set restore flag */
							regs.eflags |= 0x10000;
							break;
						}
					}
				}
			}
			correct_hw_break();
			asm volatile ( 
				      "movl %0, %%db6\n"
				      : 
				      : "r" (0) );
			goto exit_kgdb;

			/* kill the program */
		case 'k' :  /* do nothing */
			break;

			/* query */
		case 'q' :
			switch (remcomInBuffer[1]) {
			case 'L':	
				/* List threads */
				unpack_byte(remcomInBuffer+3, &maxthreads);
				unpack_threadid(remcomInBuffer+5, &thref);
				do{
					int buf_thread_limit = (BUFMAX-22)/BUF_THREAD_ID_SIZE;
					if (maxthreads > buf_thread_limit){
						maxthreads = buf_thread_limit;
					}
				}while (0);
				remcomOutBuffer[0] = 'q';
				remcomOutBuffer[1] = 'M';
				remcomOutBuffer[4] = '0';
				pack_threadid(remcomOutBuffer+5, &thref);	

				threadid = threadref_to_int(&thref);
				for (nothreads = 0;
				     nothreads < maxthreads && threadid< PID_MAX+NUM_CPUS;
				     threadid++ )
					{
						thread = getthread(threadid);
						if (thread) {
							int_to_threadref(&thref, threadid);
							pack_threadid(remcomOutBuffer+21+nothreads*16,
								      &thref);
							nothreads++;
						}
					}
			
				if (threadid == PID_MAX+NUM_CPUS) {
					remcomOutBuffer[4] = '1';
				}
				pack_hex_byte(remcomOutBuffer+2, nothreads);
				remcomOutBuffer[21+nothreads*16] = '\0';
				break;

			case 'C':
				/* Current thread id */
				remcomOutBuffer[0] = 'Q';
				remcomOutBuffer[1] = 'C';
				threadid = current->pid;
				if ( !threadid ){
					/*
					 * idle thread
					 */
					for (threadid = PID_MAX; 
					     threadid < PID_MAX+NUM_CPUS; 
					     threadid++){
						if ( current == 
                                                     idle_task(threadid - PID_MAX))
							break;
					}
				}
				int_to_threadref(&thref, threadid);
				pack_threadid(remcomOutBuffer+2, &thref);	
				remcomOutBuffer[18] = '\0';
				break;

			case 'E':
				/* Print exception info */
				printexceptioninfo(exceptionVector,
                                                   err_code, 
                                                   remcomOutBuffer);
				break;
			}
			break;

			/* task related */
		case 'H' :	 
			switch (remcomInBuffer[1]) {
			case 'g':
				ptr = &remcomInBuffer[2];
				hexToInt(&ptr, &threadid);
				thread = getthread(threadid);
				if (!thread) {
					remcomOutBuffer[0] = 'E';
					remcomOutBuffer[1] = '\0';
					break;
				}
				usethread = thread;
				/* follow through */
			case 'c':
				remcomOutBuffer[0] = 'O';
				remcomOutBuffer[1] = 'K';
				remcomOutBuffer[2] = '\0';
				break;
			}
			break;

			/* Query thread status */
		case 'T':
			ptr = &remcomInBuffer[1];
			hexToInt(&ptr, &threadid);
			thread = getthread(threadid);
			if (thread) {
				remcomOutBuffer[0] = 'O';
				remcomOutBuffer[1] = 'K';
				remcomOutBuffer[2] = '\0';
			} else {
				remcomOutBuffer[0] = 'E';
				remcomOutBuffer[1] = '\0';
			}
			break;

		case 'Y':
			ptr = &remcomInBuffer[1];
			hexToInt(&ptr, &breakno);
			ptr++;
			hexToInt(&ptr, &breaktype);
			ptr++;
			hexToInt(&ptr, &length);
			ptr++;
			hexToInt(&ptr, &addr);
			if (set_hw_break(breakno & 0x3, 
                                         breaktype & 0x3 , 
                                         length & 0x3, addr)== 0) {
				strcpy(remcomOutBuffer, "OK");
			} else {
				strcpy(remcomOutBuffer, "ERROR");
			}
			break;

			/* Remove hardware breakpoint */      
		case 'y':
			ptr = &remcomInBuffer[1];
			hexToInt(&ptr, &breakno);
			if (remove_hw_break(breakno & 0x3) == 0) {
				strcpy(remcomOutBuffer, "OK");
			} else {
				strcpy(remcomOutBuffer, "ERROR");
			}
			break;


		} /* switch */

		/* reply to the request */
		putpacket(remcomOutBuffer);
	} /* while(1==1) */
	/*
	 *  reached by goto only.
	 */
 exit_kgdb:
#ifdef CONFIG_SMP
        /*
	 * Release gdb wait locks 
         * Sanity check time.  Must have at least one cpu to run.  Also single
         * step must not be done if the current cpu is on hold.
         */
        if (spinlock_count == 1){
                int ss_hold = (regs.eflags & 0x100) &&  kgdb_info.hold_on_sstep;
                int cpu_avail = 0;
                int i;

		for (i = 0; i < smp_num_cpus; i++) {
                        if (!hold_cpu(i)){
                                cpu_avail = 1;
                        }
                }
                if ( ! cpu_avail) {
                        to_gdb("No cpus unblocked, see 'kgdb_info.hold_cpu'\n");
                        goto once_again;
                }
                if ( hold_cpu(smp_processor_id()) && 
                      (regs.eflags & 0x100)) {
                        to_gdb("Current cpu must be unblocked to single step\n");
                        goto once_again;
                }
                if (  !(ss_hold)){
                        int i;
                        for (i = 0; i < smp_num_cpus; i++) {
                                if ( ! hold_cpu(i) ){
                                        spin_unlock(&waitlocks[i]);
                                }
                        }
		}else{
                        spin_unlock(&waitlocks[smp_processor_id()]);
                }
                /* Release kgdb spinlock */
                KGDB_SPIN_UNLOCK(&kgdb_spinlock);
                /*
                 * If this cpu is on hold, this is where we
                 * do it.  Note, the NMI will pull us out of here,
                 * but will return as the above lock is not held.
                 * We will stay here till another cpu releases the lock for us.
                 */
                spin_unlock_wait(waitlocks + smp_processor_id());
                restore_flags(flags) ;
                return(0) ;
	}
 exit_just_unlock:
#endif
	/* Release kgdb spinlock */
	KGDB_SPIN_UNLOCK(&kgdb_spinlock);
	restore_flags(flags) ;
	return(0) ;
}

gdb_debug_hook *linux_debug_hook = &handle_exception;

/* this function is used to set up exception handlers for tracing and
 * breakpoints. 
 * This function is not needed as the above line does all that is needed.
 * We leave it for backward compatitability...
 */
void set_debug_traps(void)
{
	/*
	 * linux_debug_hook is defined in traps.c.  We store a pointer
	 * to our own exception handler into it.

	 * But really folks, every hear of labeled common, an old Fortran
	 * concept.  Lots of folks can reference it and it is define if 
	 * anyone does.	 Only one can initialize it at link time.  We do 
	 * this with the hook.	See the statement above.  No need for any
	 * executable code and it is ready as soon as the kernel is
	 * loaded.  Very desirable in kernel debugging.

	 linux_debug_hook = handle_exception ;
	*/

	/* In case GDB is started before us, ack any packets (presumably
	   "$?#xx") sitting there. 
	   putDebugChar ('+');

	   initialized = 1;
	*/
}

/* This function will generate a breakpoint exception.	It is used at the
   beginning of a program to sync up with a debugger and can be used
   otherwise as a quick means to stop program execution and "break" into
   the debugger. */
/* But really, just use the BREAKPOINT macro.  We will handle the int stuff
 */

#ifdef later
/*
 * possibly we should not go thru the traps.c code at all?  Someday.
 */
void do_kgdb_int3(struct pt_regs * regs, long error_code)
{
	handle_exception(3,5,error_code,regs);
	return;
}
#endif
#ifdef CONFIG_STACK_OVERFLOW_TEST
#include <asm/kgdb.h>
asmlinkage void stack_overflow(void)
{
#ifdef BREAKPOINT
	BREAKPOINT;
#else
	printk("Kernel stack overflow, looping forever\n");
#endif
	while(1) {
	}
}
#endif
#ifdef CONFIG_SMP

char	gdbconbuf[BUFMAX];

static void kgdb_gdb_message(const char *s, unsigned count)
{
	int	i;
	int	wcount;
	char	*bufptr;
        /*
         * This takes care of NMI while spining out chars to gdb
         */
        in_kgdb_console=1;
	gdbconbuf[0] = 'O';
	bufptr = gdbconbuf + 1;
	while (count > 0) {
		if ((count << 1) > (BUFMAX - 2)) {
			wcount = (BUFMAX - 2) >> 1;
		} else {
			wcount = count;
		}
		count -= wcount;
		for (i = 0; i < wcount; i++) {
			bufptr = pack_hex_byte(bufptr, s[i]);
		}
		*bufptr = '\0';
		s += wcount;

		putpacket(gdbconbuf);

	}
        in_kgdb_console=0;
}
static void to_gdb(const char *s)
{
        int count = 0;
        while ( s[count] && (count++ < BUFMAX));
        kgdb_gdb_message(s,count);
}
#endif
#ifdef CONFIG_KGDB_CONSOLE
#include <linux/console.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>


void kgdb_console_write(struct console *co, const char *s,
				unsigned count)
{

	if (gdb_i386vector == -1) {
                /*
                 * We have not yet talked to gdb.  What to do...
                 * lets break, on continue we can do the write.
                 * But first tell him whats up. Uh, well no can do,
                 * as this IS the console.  Oh well...
                 * We do need to wait or the messages will be lost.
                 * Other option would be to tell the above code to
                 * ignore this breakpoint and do an auto return, 
                 * but that might confuse gdb.  Also this happens 
                 * early enough in boot up that we don't have the traps
                 * set up yet, so...
                 */
                breakpoint();
	}
        kgdb_gdb_message(s,count);
}

static kdev_t kgdb_console_device(struct console *c);

/*
 * ------------------------------------------------------------
 * Serial KGDB driver 
 * ------------------------------------------------------------
 */

static struct console kgdbcons = {
	name:		"kgdb",
	write:          kgdb_console_write,
#ifdef CONFIG_KGDB_USER_CONSOLE
	device:		kgdb_console_device,
#endif
	flags:		CON_PRINTBUFFER | CON_ENABLED,
	index:		-1,
};

/*
 * The trick here is that this file gets linked before printk.o
 * That means we get to peer at the console info in the command
 * line before it does.  If we are up, we register, otherwise,
 * do nothing.  By returning 0, we allow printk to look also.
 */
static int kgdb_console_enabled;

int __init kgdb_console_init(char *str)
{
        if (strncmp(str,"kgdb",4) == 0){
                register_console(&kgdbcons);
                kgdb_console_enabled = 1;
        }
        return 0;         /* let others look at the string */
}
__setup("console=",kgdb_console_init);

#ifdef CONFIG_KGDB_USER_CONSOLE
/* This stuff sort of works, but it knocks out telnet devices  
 * we are leaving it here in case we (or you) find time to figure it out
 * better..
 */

/*
 * We need a real char device as well for when the console is opened for user
 * space activities.
 */

static int kgdb_consdev_open(struct inode * inode, struct file * file)
{
	return 0;
}

static ssize_t kgdb_consdev_write(struct file * file, const char *buf,
				  size_t count, loff_t *ppos)
{
	int size, ret = 0;
	static char kbuf[128];
	static DECLARE_MUTEX(sem);

	/* We are not reentrant... */
	if (down_interruptible(&sem))
		return -ERESTARTSYS;

	while (count > 0) {
		/* need to copy the data from user space */
		size = count;
		if (size > sizeof(kbuf))
			size = sizeof(kbuf);
		if (copy_from_user (kbuf, buf, size)) {
			ret = -EFAULT;
			break;;
		}
		kgdb_console_write(&kgdbcons, kbuf, size);
		count -= size;
		ret += size;
		buf += size;
	}

	up(&sem);

	return ret;
}

struct file_operations kgdb_consdev_fops = {
	open:		kgdb_consdev_open,
	write:		kgdb_consdev_write
};
static kdev_t kgdb_console_device(struct console *c)
{
	return MKDEV(TTYAUX_MAJOR, 1);
}

/*
 * This routine gets called from the serial stub in the i386/lib 
 * This is so it is done late in bring up (just before the console open).
 */
void kgdb_console_finit(void)
{
        if (kgdb_console_enabled){
                char * cptr = cdevname(MKDEV(TTYAUX_MAJOR,1));
                char * cp = cptr;
                while ( *cptr && *cptr != '(') cptr++;
                *cptr = 0;
                unregister_chrdev(TTYAUX_MAJOR,cp);
                register_chrdev(TTYAUX_MAJOR, "kgdb", &kgdb_consdev_fops);
        }
}
#endif
#endif
#ifdef CONFIG_KGDB_TS
#include <asm/msr.h> /* time stamp code */
#ifdef CONFIG_KGDB_TS_64
#define DATA_POINTS 64
#endif
#ifdef CONFIG_KGDB_TS_128
#define DATA_POINTS 128
#endif
#ifdef CONFIG_KGDB_TS_256
#define DATA_POINTS 256
#endif
#ifdef CONFIG_KGDB_TS_512
#define DATA_POINTS 512
#endif
#ifdef CONFIG_KGDB_TS_1024
#define DATA_POINTS 1024
#endif
#ifndef DATA_POINTS
#define DATA_POINTS 128	/* must be a power of two */
#endif
#define INDEX_MASK (DATA_POINTS - 1)
#if (INDEX_MASK & DATA_POINTS)
#error "CONFIG_KGDB_TS_COUNT must be a power of 2"
#endif
struct kgdb_and_then_struct {
#ifdef CONFIG_SMP
	int	on_cpu;
#endif
	long long at_time;
	int  	from_ln;
	char	* in_src;
	void	*from;
        int     with_if;
	int	data0;
	int	data1;
};
struct kgdb_and_then_struct2 {
#ifdef CONFIG_SMP
	int	on_cpu;
#endif
	long long at_time;
	int  	from_ln;
	char	* in_src;
	void	*from;
        int     with_if;
	struct task_struct *t1;
	struct task_struct *t2;
};
struct kgdb_and_then_struct kgdb_data[DATA_POINTS];

struct kgdb_and_then_struct *kgdb_and_then = &kgdb_data[0];
int kgdb_and_then_count;


void kgdb_tstamp(int line, char * source, int data0, int data1)
{
        static spinlock_t ts_spin = SPIN_LOCK_UNLOCKED;
        int flags;
        local_irq_save(flags);
        spin_lock(&ts_spin);
        rdtscll(kgdb_and_then->at_time);
#ifdef CONFIG_SMP
        kgdb_and_then->on_cpu = smp_processor_id();
#endif
        kgdb_and_then->from_ln = line;
        kgdb_and_then->in_src = source;
        kgdb_and_then->from = __builtin_return_address(0);
        kgdb_and_then->with_if = (flags & 0x200) >> 9;
        kgdb_and_then->data0 = data0;
        kgdb_and_then->data1 = data1;
        kgdb_and_then = &kgdb_data[++kgdb_and_then_count & INDEX_MASK];
        spin_unlock(&ts_spin);
        local_irq_restore(flags); 
        return;
}
#endif
