/*
 *  linux/arch/i386/snsc_mpu300/powctrl.c
 *
 *  powctrl functions for Sony NSC MPU-301 board
 *
 *  Copyright 2001,2002 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/smp_lock.h>
#include <linux/spinlock.h>
#include <linux/pci.h>
#include <linux/pci_ids.h>
#include <linux/delay.h>
#include <linux/snsc_major.h>
#include <linux/snsc_ich_gpio.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/mc146818rtc.h>
#include <asm/snsc_mpu30x_powctrl.h>

#define MODNAME            "powctrl"

/* Processor Interface Registers */
#define RST_CNT                     0xcf9
#define     FULL_RST                     (1 << 3)
#define     RST_CPU                      (1 << 2)
#define     SYS_RST                      (1 << 1)

/* ICH2 Device ID */
#define PCI_DEVICE_ID_ICH2_LPC      0x2440
#define PCI_DEVICE_ID_ICH2_PCIBRG   0x244e

/* LPC interface PCI Configuration Register */
#define PMBASE                      0x40
#define SERIRQ_CNTL                 0x64
#define     LPC_SIRQEN                   (1 << 7)

/* Power Management I/O Registers */
#define PM1_STS                     0x00
#define      WAK_STS                     (1 << 15)
#define      PWRBTN_STS                  (1 << 8)
#define      PM1_STS_MASK                (0x8d21)
#define PM1_EN                      0x02
#define      RTC_EN                      (1 << 10)
#define PM1_CNT                     0x04
#define      SLP_EN                      (1 << 13)
#define      SLP_TYP_S1                  (1 << 10)
#define      SLP_TYP_S3                  (5 << 10)
#define      SLP_TYP_S4                  (6 << 10)
#define      SLP_TYP_S5                  (7 << 10)
#define GPE0_STS                    0x28
#define      PME_STS                     (1 << 11)
#define      GPE0_STS_MASK               (0x0bfb)
#define GPE0_EN                     0x2a
#define      PME_EN                      (1 << 11)
#define      AC97_EN                     (1 << 5)
#define GPE1_STS                    0x2c
#define      GPE1_STS_MASK               (0xffdb)
#define GPE1_EN                     0x2e


/* Board ID */
#define FWH_BASE                    0xffbc0000
#define FWH_GPI                     0x100
#define     FWH_BOARDID_MPU301                (1 << 4)

/* Super I/O PME registers */
#define PME_BASE                    0x800
#define SPIO_PME_STS                (PME_BASE + 0x00)
#define SPIO_PME_EN                 (PME_BASE + 0x02)
#define SPIO_PME_STS1               (PME_BASE + 0x04)
#define SPIO_PME_STS3               (PME_BASE + 0x06)
#define SPIO_PME_STS4               (PME_BASE + 0x07)
#define SPIO_PME_EN1                (PME_BASE + 0x0a)
#define SPIO_PME_EN3                (PME_BASE + 0x0c)
#define SPIO_PME_EN4                (PME_BASE + 0x0d)
#define      PME_CIR                     (1 << 0)
#define      PME_RI2                     (1 << 1)
#define      PME_RI1                     (1 << 2)
#define      PME_KBD                     (1 << 3)
#define      PME_MOUSE                   (1 << 4)
#define      PME_SPEKEY                  (1 << 5)
#define      PME_FAN1                    (1 << 6)
#define      PME_FAN2                    (1 << 7)
#define SPIO_GP42                   (PME_BASE + 0x3d)
#define      GP_OUT                      (0 << 0)
#define      GP_IN                       (1 << 0)
#define      GP_ALTFUNC                  (1 << 2)
#define      GP_OPENDRAIN                (1 << 7)
#define      GP_PUSHPULL                 (0 << 7)

/*
 *    LAN interface (bus 1, dev 8, func 0)
 */
#define LAN_BUS                     1
#define LAN_DEV                     8
#define LAN_FUN                     0
#define LAN_PMCSR                   0xe0
#define     PMCSR_PMESTS                 (1 << 15)


/* GPIO[12]/[28] is used as Shutdown Req#/Ack */
#define GPIO_SHUTREQ                12
#define GPIO_SHUTACK                28

/* GPIO[8] is used as SuperI/O PME */
#define GPIO_LPC_PME                8

struct _mpu300_pm_info {
        struct pci_dev  *lpc;
        struct pci_dev  *pcibrg;
        u32             pm_base;
};

static struct _mpu300_pm_info  pm_info;

struct powctrl_state {
        wait_queue_head_t  waitq;
        u32                flags;
        spinlock_t         lock;
        u32                wakeup_flag;
} pwstate;

/* for pwstate.flags */
#define POWCTRL_OPENED    0x01
#define POWCTRL_REQ       0x02
#define POWCTRL_WAIT      0x04
#define POWCTRL_SHUTENA   0x08
#define POWCTRL_REQIRQ    0x10

/* for Super I/O GPIO */
#define NUM_WAKE_GPIO    10
static const struct {
	u_int8_t reg;
	u_int8_t mask;
} sgpio[NUM_WAKE_GPIO] = {
	{3, 0x02},
	{3, 0x04},
	{4, 0x10},
	{4, 0x20},
	{4, 0x40},
	{4, 0x80},
	{3, 0x01},
	{3, 0x20},
	{3, 0x40},
	{3, 0x80}};

/* Power Management event passed from NBL */
u32 pm_events;


void shutreq_handler( int irq, void *dev_id, struct pt_regs *regs)
{
        struct powctrl_state *pstat = (struct powctrl_state *)dev_id;

        if (ich_gpio_intr_get_status_bit(0, GPIO_SHUTREQ)) {
                /* mask intterupt because status cannot be cleared until signal is negated */
                ich_gpio_intr_mask_bit(0, GPIO_SHUTREQ);
                pstat->flags |= POWCTRL_REQ;
                if (waitqueue_active(&pstat->waitq))
                        wake_up_interruptible(&pstat->waitq);
                
                ich_gpio_intr_clear_bit(0, GPIO_SHUTREQ);
        }

        return;
}


/*
 * restart 
 */
void _mpu300_machine_restart(char * __unused)
{
    /* use RST_CNT */
    outb_p(RST_CPU | SYS_RST, RST_CNT);
}


/*
 *  disable all wakeup event
 */
static inline void
_maskall_wakeup(void)
{
        /* disable PM1, GPE0, GPE1 */
        outw_p(0, pm_info.pm_base + PM1_EN);
	outw_p(0, pm_info.pm_base + GPE0_EN);
	outw_p(0, pm_info.pm_base + GPE1_EN);
}


/*
 *  clear all wakeup event status
 */
static inline void
_clearall_wakeup(void)
{
	int i;
	struct pci_dev *lan_dev;
	u_int8_t v8;

	/*
	 *  Super I/O
	 */
	/* clear PME status */
	for (i = 0; i < 5; i++) {
		outb_p(inb_p(SPIO_PME_STS1 + i), SPIO_PME_STS1 + i);
	}
	if (inb_p(SPIO_PME_STS) & 1) {
		outb_p(1, SPIO_PME_STS);
	}

	/*
	 *  ICH2
	 */

	/* clear RTC interrupts */
	v8 = CMOS_READ(RTC_INTR_FLAGS);

        /* clear PM1_STS */
        outw_p(inw_p(pm_info.pm_base + PM1_STS) & PM1_STS_MASK,
	       pm_info.pm_base + PM1_STS);

	/* clear PME status of LAN controller */
	lan_dev = pci_find_slot (LAN_BUS, PCI_DEVFN(LAN_DEV, LAN_FUN));
	if (lan_dev != NULL) {
		u16 val;
		if ((pci_read_config_word(lan_dev, LAN_PMCSR, &val) == 0) && 
		    (val & PMCSR_PMESTS))
			pci_write_config_word(lan_dev, LAN_PMCSR,
					      val | PMCSR_PMESTS);
	}

	/* clear GPE0, GPE1 status */
	outw_p(inw_p(pm_info.pm_base + GPE0_STS) & GPE0_STS_MASK,
	       pm_info.pm_base + GPE0_STS);
	outw_p(inw_p(pm_info.pm_base + GPE1_STS) & GPE1_STS_MASK,
	       pm_info.pm_base + GPE1_STS);
}


/* 
 * power off 
 */
void _mpu300_power_off(void)
{
	int i;
        u8  v8;
        struct powctrl_state *pstat = &pwstate;

	local_irq_disable();

        /* disable Serial IRQ */
        pci_read_config_byte(pm_info.lpc, SERIRQ_CNTL, &v8);
        pci_write_config_byte(pm_info.lpc, SERIRQ_CNTL, v8 & ~LPC_SIRQEN);

        /* disable PCI master */
        pci_read_config_byte(pm_info.pcibrg, PCI_COMMAND, &v8);
        pci_write_config_byte(pm_info.pcibrg, PCI_COMMAND, 
                              v8 & ~PCI_COMMAND_MASTER);

        /* mask Shutdown_Req# interrupt, because wake_event_interruptible()
           in powerctl_ioctl() is not ended and the intterupt is not disabled
           when force power-off is issued(e.g. # halt -f -p) */
        ich_gpio_intr_mask_bit(0, GPIO_SHUTREQ);

	/* enable nIO_PME output (active low) */
	outb_p(GP_OUT | GP_ALTFUNC, SPIO_GP42);


	/* mask all wakeup events */
	_maskall_wakeup();

	/*
	 *  enable wake events
	 */
	/* any Super I/O events */
	if (pstat->wakeup_flag 
	    & (WAKEUP_RI0 | WAKEUP_RI1 | WAKEUP_SGPIO_ALL)) {
		/* Super I/O PME enable */
		outb_p(1, SPIO_PME_EN);
		
                /* enable GPIO[8](= LPC_PME) */
                ich_gpio_intr_mask_bit(0, GPIO_LPC_PME);
		ich_gpio_register_bit(0, GPIO_LPC_PME, "mpu301_powctrl");
		ich_gpio_set_mode_bit(0, GPIO_LPC_PME, GPIO_MODE_IN);
		ich_gpio_intr_set_mode_bit(0, GPIO_LPC_PME,
					   GPIO_INTR_LEVEL_LOW |
 					   GPIO_INTR_ROUT_SCI);
                ich_gpio_intr_clear_bit(0, GPIO_LPC_PME);
                ich_gpio_intr_unmask_bit(0,GPIO_LPC_PME);
	}

	/* WOL */
        if (pstat->wakeup_flag & WAKEUP_WOL) {
                outw_p(PME_EN | inw_p(pm_info.pm_base + GPE0_EN),
		       pm_info.pm_base + GPE0_EN);
        }

	/* RI0 */
	if (pstat->wakeup_flag & WAKEUP_RI0) {
                outb_p(PME_RI1 | inb_p(SPIO_PME_EN1), SPIO_PME_EN1);
	}

	/* RI1 */
	if (pstat->wakeup_flag & WAKEUP_RI1) {
                outb_p(PME_RI2 | inb_p(SPIO_PME_EN1), SPIO_PME_EN1);
	}

	/* RTC */
	if (pstat->wakeup_flag & WAKEUP_RTC) {
		CMOS_WRITE(CMOS_READ(RTC_CONTROL) | RTC_AIE, RTC_CONTROL);
		outw_p(RTC_EN | inw_p(pm_info.pm_base + PM1_EN),
		       pm_info.pm_base + PM1_EN);
	}

#if 0
	/* AC97 */
	if (pstat->wakeup_flag & WAKEUP_AC97) {
                outw_p(AC97_EN | inw_p(pm_info.pm_base + GPE0_EN),
		       pm_info.pm_base + GPE0_EN);		
	}
#endif

	/* SMB */
	if (pstat->wakeup_flag & WAKEUP_SMB) {
                /* enable GPIO[11] */
                ich_gpio_intr_mask_bit(0, 11);
		ich_gpio_register_bit(0, 11, "mpu301_powctrl");
		ich_gpio_set_mode_bit(0, 11, GPIO_MODE_IN);
		ich_gpio_intr_set_mode_bit(0, 11,
					   GPIO_INTR_LEVEL_LOW |
					   GPIO_INTR_ROUT_SCI);
		ich_gpio_intr_clear_bit(0, 11);
		ich_gpio_intr_unmask_bit(0, 11);
	}

	/* Super I/O GPIO */
	if (pstat->wakeup_flag & WAKEUP_SGPIO_ALL) {
		for (i = 0; i < NUM_WAKE_GPIO; i++) {
			if (pstat->wakeup_flag & (WAKEUP_SGPIO0 << i)) {
				outb_p(sgpio[i].mask 
				       | inb_p(SPIO_PME_EN1 + sgpio[i].reg - 1),
				       SPIO_PME_EN1 + sgpio[i].reg - 1);
			}
		}
	}


	/* clear all wakeup events */
	_clearall_wakeup();

        if (pstat->flags & POWCTRL_SHUTENA) {
                /* assert Shutdown Ack */
		ich_gpio_set_data_bit(0, GPIO_SHUTACK, 1);
        }

	/* S5 */
        outl_p(SLP_EN | SLP_TYP_S5, pm_info.pm_base + PM1_CNT);

        /* wait until sleep state */
        do {
                printk("WAK_STS = 0x%x\n", inw_p(pm_info.pm_base + PM1_STS));
                udelay(10000);
        } while (!(inw_p(pm_info.pm_base + PM1_STS) & WAK_STS));
}


#define CHK_SHUTREQ(pstat) \
({ \
        if (ich_gpio_intr_get_status_bit(0, GPIO_SHUTREQ)) { \
                pstat->flags |= POWCTRL_REQ; \
        } \
        (pstat->flags & POWCTRL_REQ); \
})
static int powctrl_ioctl(struct inode *inode, struct file *file,
                         unsigned int cmd, unsigned long arg)
{
        int           rval;
        unsigned long flags;
        struct powctrl_state *pstat = &pwstate;

        if (MINOR(inode->i_rdev) != 0) {
                printk(__FUNCTION__ ": invalid minor number %d\n", MINOR(inode->i_rdev));
                return -ENODEV;
        }

        if (!(pstat->flags & POWCTRL_OPENED))
                return -EBUSY;

        switch (cmd) {
        case SNSC_POWCTRL_IOC_SHUTDOWN:
                spin_lock_irqsave(&pstat->lock ,flags);
                if (pstat->flags & POWCTRL_WAIT) {
                        /* Only one process can wait shutreq */
                        spin_unlock_irqrestore(&pstat->lock, flags);
                        return -EBUSY;
                }
                pstat->flags |= POWCTRL_WAIT;
                spin_unlock_irqrestore(&pstat->lock, flags);

                if (!(pstat->flags & POWCTRL_REQIRQ)) {
                        if (ich_gpio_intr_register_handler(0, GPIO_SHUTREQ, 
							   shutreq_handler,
							   &pwstate)) {
                                printk(KERN_ERR "mpu301 powctrl driver aborting\n");
                                return -EBUSY;
                        }
                        pstat->flags |= POWCTRL_REQIRQ;
                }

                ich_gpio_intr_unmask_bit(0, GPIO_SHUTREQ);
                rval = wait_event_interruptible(pstat->waitq, CHK_SHUTREQ(pstat));
                ich_gpio_intr_mask_bit(0, GPIO_SHUTREQ);

                if (rval != 0) {
                        /* if wait event is finished with error(ERESTARTSYS), 
                           wait event can be retried */
                        spin_lock_irqsave(&pstat->lock ,flags);
                        pstat->flags &= ~POWCTRL_WAIT;
                        spin_unlock_irqrestore(&pstat->lock, flags);
                }

                return rval;

        case SNSC_POWCTRL_IOC_SET_WAKEUP:
		pstat->wakeup_flag |= arg;
                break;

        case SNSC_POWCTRL_IOC_UNSET_WAKEUP:
		pstat->wakeup_flag &= ~arg;
                break;

        case SNSC_POWCTRL_IOC_GET_SETTINGS:
		copy_to_user((void *)arg, &pstat->wakeup_flag, 
			     sizeof(u_int32_t));
		break;

        case SNSC_POWCTRL_IOC_GET_WUPCAUSE:
		if (pm_events == 0xffffffff)
			return -EINVAL;
		copy_to_user((void *)arg, &pm_events, sizeof(u_int32_t));
		break;

        default:
                return -ENOIOCTLCMD;
        }

        return 0;
}


static int powctrl_open(struct inode *inode, struct file *file)
{
        struct powctrl_state *pstat = &pwstate;

        if (MINOR(inode->i_rdev) != 0) {
                printk(__FUNCTION__ ": invalid minor number %d\n", MINOR(inode->i_rdev));
                return -ENODEV;
        }

        pstat->flags |= POWCTRL_OPENED;

        MOD_INC_USE_COUNT;

        return 0;
}

static int powctrl_release(struct inode *inode, struct file *file)
{
        MOD_DEC_USE_COUNT;
        return 0;
}


struct file_operations powctrl_fops = {
	ioctl:		powctrl_ioctl,
	open:		powctrl_open,
	release:	powctrl_release,
};

static int bid_mpu301(void)
{
        u32  fwh_base;
        u8   gpi;

        fwh_base = (u32)ioremap_nocache(FWH_BASE, PAGE_SIZE);

        gpi = readb(fwh_base + FWH_GPI);
        if (gpi & FWH_BOARDID_MPU301) {
                return 1;
        }
        return 0;
}

static int __init _mpu300_pm_init(void)
{
        u32    g;
        int    res;
        struct powctrl_state *pstat = &pwstate;

        memset(pstat, 0, sizeof(struct powctrl_state));

        /* find LPC */
        pm_info.lpc = pci_find_device (PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_ICH2_LPC, NULL);
        if (!pm_info.lpc) {
                printk("mpu300_pm_init: cannot find %x\n", PCI_DEVICE_ID_ICH2_LPC);
                return 1;
        }

        /* find PCI Bridge */
        pm_info.pcibrg = pci_find_device (PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_ICH2_PCIBRG, NULL);
        if (!pm_info.pcibrg) {
                printk("mpu300_pm_init: cannot find %x\n", PCI_DEVICE_ID_ICH2_PCIBRG);
                return 1;
        }

        /* PM base */
        pci_read_config_dword (pm_info.lpc, PMBASE, &pm_info.pm_base);
        pm_info.pm_base &= PCI_BASE_ADDRESS_IO_MASK;
        printk("mpu300_pm_init: PM base = 0x%x\n", pm_info.pm_base);

        pm_power_off = _mpu300_power_off;

        if (bid_mpu301()) {
                if (ichgpio_init()) {
                        return 1;
                }

                /* set ICH2 GPIO[12] as LOW Active, SCI */
		ich_gpio_register_bit(0, GPIO_SHUTREQ, "mpu301_powctrl");
		ich_gpio_set_mode_bit(0, GPIO_SHUTREQ, GPIO_MODE_IN);
		ich_gpio_intr_set_mode_bit(0, GPIO_SHUTREQ,
					   GPIO_INTR_LEVEL_LOW |
					   GPIO_INTR_ROUT_SCI);
                ich_gpio_intr_mask_bit(0, GPIO_SHUTREQ);
                ich_gpio_intr_clear_bit(0, GPIO_SHUTREQ);

                /* set ICH2 GPIO[28] as Output */
		ich_gpio_register_bit(0, GPIO_SHUTACK, "mpu301_powctrl");
		ich_gpio_set_mode_bit(0, GPIO_SHUTACK, GPIO_MODE_OUT_HIGH);
        
                /* negate Shutdown Ack 
                   and acceptable Shutdown Req# */
                g = ich_gpio_get_data_bit(0, GPIO_SHUTACK);
		ich_gpio_set_data_bit(0, GPIO_SHUTACK, !g);

                pstat->flags = POWCTRL_SHUTENA;
                spin_lock_init(&pstat->lock);
                init_waitqueue_head(&pstat->waitq);

                if ((res = register_chrdev(SNSC_POWCTRL_MAJOR, MODNAME, &powctrl_fops)) < 0) {
                        printk(__FUNCTION__ ": cannot register driver\n");
                        return res;
                }

		/* mask all wakeup events */
		_maskall_wakeup();

                printk("MPU-301 powctrl driver $Revision: 1.1.1.1 $\n");
        }
        return 0;
}

static void __exit _mpu300_pm_exit(void)
{
        if (bid_mpu301()) {
                unregister_chrdev(SNSC_POWCTRL_MAJOR, MODNAME);
		ich_gpio_intr_unregister_handler(0, GPIO_SHUTREQ, &pwstate);
        }
}

module_init(_mpu300_pm_init);
module_exit(_mpu300_pm_exit);
