/* arch/mips/ite-boards/qed-4n-s01b/qed_vst.c
 * VST support for QED5231 board
 * 
 * Copyright (C) 2003 Sony Corporation 
 * Copyright (C) 2003 Matsushita Electric Industrial Co., Ltd.
 * Copyright (C) 2003 MontaVista Software, Inc.
 * Author: MontaVista Software, Inc.
 *     source@mvista.com
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/timer.h>
#include <linux/irq.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <asm/it8172/it8172.h>
#include <asm/it8172/it8172_int.h>
#include <asm/it8172/pm.h>

#define PREFIX "IT8172 (QED5231) VST: "
#undef DEBUG_VST
#ifdef DEBUG_VST
#define DPRINT(args...) printk(PREFIX args)
#else
#define DPRINT(args...) do {} while (0)
#endif

/* Timer registers access */
#define IT8172_TIMER_BASE           (IT8172_PCI_IO_BASE + IT_TIMER_BASE) /* 0x14010800 */
#define IT8172_TIMER_READ(reg)      *(volatile u16 *)KSEG1ADDR((IT8172_TIMER_BASE + (reg)))
#define IT8172_TIMER_WRITE(reg,val) *(volatile u16 *)KSEG1ADDR((IT8172_TIMER_BASE + (reg))) = (val)

static volatile unsigned short it8172_intc_lb_mask, it8172_intc_lpc_mask,
	                       it8172_intc_pci_mask, it8172_intc_nmi_mask;

static int watchdog_counter, watchdog_ints, watchdog_handled;
static int timer_counter, timer_ints, timer_handled;
static int update_vst_jiffies;

idle_state_t vst_idle_state;

/* Stuff to save/restore INTC state */
static volatile unsigned short it8172_intc_lb_mask, it8172_intc_lpc_mask,
	                       it8172_intc_pci_mask, it8172_intc_nmi_mask;

/* Save IRQ controller state and disable CPU timer before entering idle state. Note TIMER0,
   serial and watchdog irqs should be unmasked in lb_mask and nmi_mask here */
void it8172_intc_save(void)
{
	if (vst_idle_state == not_idle) {
		it8172_intc_lb_mask = it8172_hw0_icregs->lb_mask;
		it8172_intc_lpc_mask = it8172_hw0_icregs->lpc_mask;
		it8172_intc_pci_mask = it8172_hw0_icregs->pci_mask;
		it8172_intc_nmi_mask = it8172_hw0_icregs->nmi_mask;
		it8172_hw0_icregs->lb_mask = 0xff7d;
		it8172_hw0_icregs->lpc_mask = 0xffff;
		it8172_hw0_icregs->pci_mask = 0xffff;
		it8172_hw0_icregs->nmi_mask = 0xffbf;
		clear_c0_status(IE_IRQ5);
		vst_idle_state = idle;
	}
}

/* Restore IRQ constoller state and enable CPU timer */
void it8172_intc_restore(void)
{
	if (vst_idle_state == idle) {
		set_c0_status(IE_IRQ5);
		it8172_hw0_icregs->lb_mask = it8172_intc_lb_mask;
		it8172_hw0_icregs->lpc_mask = it8172_intc_lpc_mask;
		it8172_hw0_icregs->pci_mask = it8172_intc_pci_mask;
		it8172_hw0_icregs->nmi_mask = it8172_intc_nmi_mask;
		vst_idle_state = not_idle;
	}
}

/* Update jiffies after sleeping for a while */
void inline update_jiffies_vst(void)
{
	u64 update = 1;

	if (update_vst_jiffies) {
		if (watchdog_ints && timer_ints)
			update += ((watchdog_ints * HZ) + timer_ints);
		watchdog_ints = timer_ints = 0;
	}
	vst_idle_state = not_idle;
	(*(u64 *)&jiffies_64) += update;
	return;
}

static void qed_timer_irq(int irq, void *unused, struct pt_regs *regs)
{
	/* Sanity check we have irq from our timer */
	if (IT8172_TIMER_READ(TIMER_TIRR) & 0x1) {
		timer_counter++;
		if (timer_counter == timer_ints) {
			/* Stop the timer */
			IT8172_TIMER_WRITE(TIMER_TCR0, 0x0);
			/* Free IRQ */
			free_irq(IT8172_TIMER0_IRQ, NULL);
			timer_handled = 0;
			/* Re-enable ticks */
			IT8172_RESTORE();
			timer_counter = 0;
			vst_idle_state = not_idle;
		}
		vst_idle_state = idle;
	}
}

static int qed_run_timer(void)
{
	int ret = 0;
	if ((ret = request_irq(IT8172_TIMER0_IRQ, qed_timer_irq, 0, "Timer 0", NULL))) {
		printk(KERN_ERR PREFIX "can't request timer 0 IRQ%d, error %d\n",
		       IT8172_TIMER0_IRQ, ret);
		return ret;
	}
	timer_handled = 1;

	/* Set timer constant value */
	IT8172_TIMER_WRITE(TIMER_TCVR0, 0x0);
	
	/* Run the timer at min. possible (CPU/16) clock frequency */
	IT8172_TIMER_WRITE(TIMER_TCR0, 0x1);
	return ret;
}

static void qed_watchdog_irq(int irq, void *dev_id, struct pt_regs *regs)
{
	watchdog_counter++;
	if (watchdog_counter == watchdog_ints) {
		/* Stop the watchdog timer */
		IT8172_TIMER_WRITE(TIMER_WDCTRL, 0x0);
		/* Free watchdog IRQ */
		free_irq(IT8172_WDOG_NMI_IRQ, &watchdog_counter);
		watchdog_handled = 0;
		if (timer_ints) {
			if (qed_run_timer())
				printk(KERN_ERR PREFIX "can't run timer\n");
		}
		else {
			/* Re-enable ticks */
			DPRINT("re-enable ticks from watchdog\n");
			IT8172_RESTORE();
			watchdog_counter = 0;
		}
	}
}

static int qed_vst(void)
{
	int ret = 0;

	if (watchdog_ints) {
		if ((ret = request_irq(IT8172_WDOG_NMI_IRQ, qed_watchdog_irq,
				       SA_SHIRQ, "it8172_watchdog", &watchdog_counter))) {
			printk(KERN_ERR PREFIX "can't request watchdog IRQ %d, error %d\n",
			       IT8172_WDOG_NMI_IRQ, ret);
			return ret;
		}
		watchdog_handled = 1;
		/* Configure watchdog to generate INT#0 instead of reset and enable it */
		IT8172_TIMER_WRITE(TIMER_WDCTRL, 0x1);
	} else
		if (qed_run_timer())
			printk(KERN_ERR PREFIX "can't run timer\n");
	return ret;
}

/* This is the main idle function called whenever we're going into idle state */
void vst_idle(void)
{
	extern void timer_bh(void);
	unsigned long next_timer, flags;
	unsigned long j = 1;

	spin_lock_irqsave(&timerlist_lock, flags);
	switch (find_next_timer(&next_timer)) {
	    case SHORT_NEXT_TIMER:
		/* Next timer is less than threshold; leave the periodic tick
		   interrupt enabled and don't set up a non-periodic timer */
		vst_short_timers++;
		DPRINT("Short timer %lu\n", next_timer);
		spin_unlock_irqrestore(&timerlist_lock, flags);
		break;
	    case NO_NEXT_TIMER:
		/* No timers in lists; If we didn't have to worry about
		   timer overflow, we could disable the periodic tick
		   interrupt and don't bother with a non-periodic timer.
		   We would then wait for an interrupt to wake us up.
		   However, that is not the case.  next_timer has been
		   set to the maximum interval, so just fall through and
		   handle it the same as RETURNED_NEXT_TIMER */
		vst_no_timers++;
		/* FALLTHROUGH ! */
	    case RETURNED_NEXT_TIMER:
	    	if (next_timer < vst_threshold)
			break;
		if (next_timer == MAX_TIMER_INTERVAL) {
			/* 1s sleep with watchdog timer only */
			watchdog_ints = 1;
			timer_ints = 0;
		} else {
			j = jiffies_64;
			watchdog_ints = next_timer/HZ;
			timer_ints = next_timer % HZ;
			/*
			 * I don't think so.
			 * timer_ints = (next_timer - j) - watchdog_ints * HZ;
			 */
		}
		update_vst_jiffies = 1;
		vst_long_timers++;
		watchdog_counter = timer_counter = 0;
		if (qed_vst()) {
			goto no_vst;
		}
		IT8172_SAVE();
		do {
			/* Note cpu_wait is not NULL for this arch */
			cpu_wait();
		} while (vst_idle_state == idle);
		IT8172_RESTORE();
		/*
		 * we are going to update jiffies here.  The jiffie
		 * clock is the only function that calls do_timer.
		 * It updates process and system accounting.
		 */ 
		update_jiffies_vst(); 
		mark_bh(TIMER_BH);
		if (TQ_ACTIVE(tq_timer))
			mark_bh(TQUEUE_BH);
		timer_bh();
		if (timer_handled) {
			free_irq(IT8172_TIMER0_IRQ, NULL);
			timer_handled = 0;
		} else 
			DPRINT("didn't free timer irq\n");
		if (watchdog_handled) {
			free_irq(IT8172_WDOG_NMI_IRQ, &watchdog_counter);
			watchdog_handled = 0;
		} else
			DPRINT("didn't free watchdog irq\n");
	no_vst:
		spin_unlock_irqrestore(&timerlist_lock, flags);
		if (need_resched())
			schedule();
	}
}

static int __init qed_vst_init(void)
{
	unsigned short it_dsr;

	/* Set normal operation mode for timer 0 (turn STANDBY off).
	   FIXME: should be in it8172_setup() instead ? */
	IT_IO_READ16(IT_PM_DSR, it_dsr);
	it_dsr &= ~IT_PM_DSR_TMR0SB;
	IT_IO_WRITE16(IT_PM_DSR, it_dsr);
	
	/* Configure watchdog timer */
	it8172_hw0_icregs->nmi_redir |= 0x0040; /* Redirect NMI# from watchdog to INT0# */
	IT8172_TIMER_WRITE(TIMER_WDMR, 0x1);    /* 1 to match with watchdog counter */

	vst_idle_state = not_idle;
	pm_idle = vst_idle;
	printk(KERN_INFO PREFIX "initialized pm_idle at 0x%x\n", pm_idle);
	return 0;
}

__initcall(qed_vst_init);
