/*
 *  /proc/tx4927_reg to read TX4927's registers.
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#include <asm-mips/addrspace.h>
#include <asm-mips/tx4927/tx4927.h>

/* ======================================================================= *
 * Constant                                                                *
 * ======================================================================= */

#define TX4927_REG_PROC_DEVNAME		"tx4927_reg_proc"
#define TX4927_REG_PROC_DIRNAME		"tx4927_reg"
#define TX4927_REG_PROC_MODULE_FILENAME	"all"

#define TX4927_REG_PROC_MAX_NAMELEN	32

MODULE_AUTHOR("Sony PTC");
MODULE_DESCRIPTION("TX4927 register proc filesystem");
MODULE_LICENSE("GPL");

/* ======================================================================= *
 * Debug print function from dpr package                                   *
 * ======================================================================= */

#define DPL_EMERG	0
#define DPL_ALERT	1
#define DPL_CRIT	2
#define DPL_ERR		3
#define DPL_WARNING	4
#define DPL_NOTICE	5
#define DPL_INFO	6
#define DPL_DEBUG	7

#define DPM_PORPOSE	0x0000000f
#define DPM_ERR		0x00000001
#define DPM_TRACE	0x00000002
#define DPM_LAYER	0x000000f0
#define DPM_APPLICATION	0x00000010
#define DPM_MIDDLEWARE	0x00000020
#define DPM_DRIVER	0x00000040
#define DPM_KERNEL	0x00000080
#define DPM_USERDEF	0xffff0000

#ifdef DEBUG

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DPL_DEBUG
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  DPM_PORPOSE
#endif

#ifndef DPR
#include <stdio.h>
#define DPR(fmt, args...)	fprintf(stderr, fmt, ## args)
#endif

#define DPRL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPR(fmt, ## args))
#define DPRM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPR(fmt, ## args))
#define DPRLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPR(fmt, ## args))
#define DPRC(cond, fmt, args...) \
	(void)((cond) && DPR(fmt, ## args))
#define DPRF(fmt, args...)	DPR("%s: " fmt, __FUNCTION__, ## args)
#define DPRFL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPRF(fmt, ## args))
#define DPRFM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPRF(fmt, ## args))
#define DPRFLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPRF(fmt, ## args))
#define DPRFC(cond, fmt, args...) \
	(void)((cond) && DPRF(fmt, ## args))

#define	DPRFS	DPRFLM(DPL_DEBUG, DPM_TRACE, "start\n")
#define	DPRFE	DPRFLM(DPL_DEBUG, DPM_TRACE, "end\n")
#define DPRFR	DPRFLM(DPL_DEBUG, DPM_TRACE, "return\n")

#else  /* DEBUG */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL -1
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  0
#endif

#ifndef DPR
#define DPR(fmt, args...)
#endif

#define DPRL(level, fmt, args...)
#define DPRM(mask, fmt, args...)
#define DPRLM(level, mask, fmt, args...)
#define DPRC(cond, fmt, args...)
#define DPRF(fmt, args...)
#define DPRFL(level, fmt, args...)
#define DPRFM(mask, fmt, args...)
#define DPRFLM(level, mask, fmt, args...)
#define DPRFC(cond, fmt, args...)

#define	DPRFS
#define	DPRFE
#define DPRFR

#endif /* DEBUG */

/* ======================================================================= *
 * Type and structure                                                      *
 * ======================================================================= */

typedef struct tx4927_reg_proc_reg_t {
	const char	*name;
	int		 bitwidth;
	unsigned long	 offset;
	unsigned short	 low_ino;
} tx4927_reg_proc_reg_t;

typedef struct tx4927_reg_proc_module_t {
	const char			*name;
	tx4927_reg_proc_reg_t		*regs;
	struct proc_dir_entry		*dir;
	unsigned short			 low_ino;
} tx4927_reg_proc_module_t;

/* ======================================================================= *
 * Prototype declaration                                                   *
 * ======================================================================= */

static int tx4927_reg_proc_init(void);
static void tx4927_reg_proc_exit(void);
static ssize_t tx4927_reg_proc_module_read(struct file *filp, char *buf,
					   size_t count, loff_t *f_pos);
static ssize_t tx4927_reg_proc_reg_read(struct file *filp, char *buf,
					size_t count, loff_t *f_pos);

/* ======================================================================= *
 * Global variable                                                         *
 * ======================================================================= */

static struct proc_dir_entry *tx4927_reg_proc_root;

static tx4927_reg_proc_reg_t tx4927_reg_proc_sdramc_regs[] =
{
	{"SDCCR0",	64,	TX4927_SDRAMC_SDCCR0},
	{"SDCCR1",	64,	TX4927_SDRAMC_SDCCR1},
	{"SDCCR2",	64,	TX4927_SDRAMC_SDCCR2},
	{"SDCCR3",	64,	TX4927_SDRAMC_SDCCR3},
	{"SDCTR",	64,	TX4927_SDRAMC_SDCTR},
	{"SDCCMD",	64,	TX4927_SDRAMC_SDCMD},
	{"SDCSFCMD",	64,	TX4927_SDRAMC_SDCSFCMD},
	{NULL}
};

static tx4927_reg_proc_reg_t tx4927_reg_proc_ebusc_regs[] =
{
	{"EBCCR0",	64,	TX4927_EBUSC_EBCCR0},
	{"EBCCR1",	64,	TX4927_EBUSC_EBCCR1},
	{"EBCCR2",	64,	TX4927_EBUSC_EBCCR2},
	{"EBCCR3",	64,	TX4927_EBUSC_EBCCR3},
	{"EBCCR4",	64,	TX4927_EBUSC_EBCCR4},
	{"EBCCR5",	64,	TX4927_EBUSC_EBCCR5},
	{"EBCCR6",	64,	TX4927_EBUSC_EBCCR6},
	{"EBCCR7",	64,	TX4927_EBUSC_EBCCR7},
	{NULL}
};

static tx4927_reg_proc_reg_t tx4927_reg_proc_ecc_regs[] =
{
	{"ECCCR",	64,	TX4927_ECC_ECCCR},
	{"ECCSR",	64,	TX4927_ECC_ECCSR},
	{NULL}
};

static tx4927_reg_proc_reg_t tx4927_reg_proc_dmac0_regs[] =
{
	{"DM0CHAR0",	64,	TX4927_DMAC_DMCHAR(0, 0)},
	{"DM0SAR0",	64,	TX4927_DMAC_DMSAR(0, 0)},
	{"DM0DAR0",	64,	TX4927_DMAC_DMDAR(0, 0)},
	{"DM0CNTR0",	64,	TX4927_DMAC_DMCNTR(0, 0)},
	{"DM0SAIR0",	64,	TX4927_DMAC_DMSAIR(0, 0)},
	{"DM0DAIR0",	64,	TX4927_DMAC_DMDAIR(0, 0)},
	{"DM0CCR0",	64,	TX4927_DMAC_DMCCR(0, 0)},
	{"DM0CSR0",	64,	TX4927_DMAC_DMCSR(0, 0)},

	{"DM0CHAR1",	64,	TX4927_DMAC_DMCHAR(0, 1)},
	{"DM0SAR1",	64,	TX4927_DMAC_DMSAR(0, 1)},
	{"DM0DAR1",	64,	TX4927_DMAC_DMDAR(0, 1)},
	{"DM0CNTR1",	64,	TX4927_DMAC_DMCNTR(0, 1)},
	{"DM0SAIR1",	64,	TX4927_DMAC_DMSAIR(0, 1)},
	{"DM0DAIR1",	64,	TX4927_DMAC_DMDAIR(0, 1)},
	{"DM0CCR1",	64,	TX4927_DMAC_DMCCR(0, 1)},
	{"DM0CSR1",	64,	TX4927_DMAC_DMCSR(0, 1)},

	{"DM0CHAR2",	64,	TX4927_DMAC_DMCHAR(0, 2)},
	{"DM0SAR2",	64,	TX4927_DMAC_DMSAR(0, 2)},
	{"DM0DAR2",	64,	TX4927_DMAC_DMDAR(0, 2)},
	{"DM0CNTR2",	64,	TX4927_DMAC_DMCNTR(0, 2)},
	{"DM0SAIR2",	64,	TX4927_DMAC_DMSAIR(0, 2)},
	{"DM0DAIR2",	64,	TX4927_DMAC_DMDAIR(0, 2)},
	{"DM0CCR2",	64,	TX4927_DMAC_DMCCR(0, 2)},
	{"DM0CSR2",	64,	TX4927_DMAC_DMCSR(0, 2)},

	{"DM0CHAR3",	64,	TX4927_DMAC_DMCHAR(0, 3)},
	{"DM0SAR3",	64,	TX4927_DMAC_DMSAR(0, 3)},
	{"DM0DAR3",	64,	TX4927_DMAC_DMDAR(0, 3)},
	{"DM0CNTR3",	64,	TX4927_DMAC_DMCNTR(0, 3)},
	{"DM0SAIR3",	64,	TX4927_DMAC_DMSAIR(0, 3)},
	{"DM0DAIR3",	64,	TX4927_DMAC_DMDAIR(0, 3)},
	{"DM0CCR3",	64,	TX4927_DMAC_DMCCR(0, 3)},
	{"DM0CSR3",	64,	TX4927_DMAC_DMCSR(0, 3)},

	{"DM0MFDR",	64,	TX4927_DMAC_DMMFDR(0)},
	{"DM0MCR",	64,	TX4927_DMAC_DMMCR(0)},
	{NULL}
};

static tx4927_reg_proc_reg_t tx4927_reg_proc_dmac1_regs[] =
{
	{"DM1CHAR0",	64,	TX4927_DMAC_DMCHAR(1, 0)},
	{"DM1SAR0",	64,	TX4927_DMAC_DMSAR(1, 0)},
	{"DM1DAR0",	64,	TX4927_DMAC_DMDAR(1, 0)},
	{"DM1CNTR0",	64,	TX4927_DMAC_DMCNTR(1, 0)},
	{"DM1SAIR0",	64,	TX4927_DMAC_DMSAIR(1, 0)},
	{"DM1DAIR0",	64,	TX4927_DMAC_DMDAIR(1, 0)},
	{"DM1CCR0",	64,	TX4927_DMAC_DMCCR(1, 0)},
	{"DM1CSR0",	64,	TX4927_DMAC_DMCSR(1, 0)},

	{"DM1CHAR1",	64,	TX4927_DMAC_DMCHAR(1, 1)},
	{"DM1SAR1",	64,	TX4927_DMAC_DMSAR(1, 1)},
	{"DM1DAR1",	64,	TX4927_DMAC_DMDAR(1, 1)},
	{"DM1CNTR1",	64,	TX4927_DMAC_DMCNTR(1, 1)},
	{"DM1SAIR1",	64,	TX4927_DMAC_DMSAIR(1, 1)},
	{"DM1DAIR1",	64,	TX4927_DMAC_DMDAIR(1, 1)},
	{"DM1CCR1",	64,	TX4927_DMAC_DMCCR(1, 1)},
	{"DM1CSR1",	64,	TX4927_DMAC_DMCSR(1, 1)},

	{"DM1CHAR2",	64,	TX4927_DMAC_DMCHAR(1, 2)},
	{"DM1SAR2",	64,	TX4927_DMAC_DMSAR(1, 2)},
	{"DM1DAR2",	64,	TX4927_DMAC_DMDAR(1, 2)},
	{"DM1CNTR2",	64,	TX4927_DMAC_DMCNTR(1, 2)},
	{"DM1SAIR2",	64,	TX4927_DMAC_DMSAIR(1, 2)},
	{"DM1DAIR2",	64,	TX4927_DMAC_DMDAIR(1, 2)},
	{"DM1CCR2",	64,	TX4927_DMAC_DMCCR(1, 2)},
	{"DM1CSR2",	64,	TX4927_DMAC_DMCSR(1, 2)},

	{"DM1CHAR3",	64,	TX4927_DMAC_DMCHAR(1, 3)},
	{"DM1SAR3",	64,	TX4927_DMAC_DMSAR(1, 3)},
	{"DM1DAR3",	64,	TX4927_DMAC_DMDAR(1, 3)},
	{"DM1CNTR3",	64,	TX4927_DMAC_DMCNTR(1, 3)},
	{"DM1SAIR3",	64,	TX4927_DMAC_DMSAIR(1, 3)},
	{"DM1DAIR3",	64,	TX4927_DMAC_DMDAIR(1, 3)},
	{"DM1CCR3",	64,	TX4927_DMAC_DMCCR(1, 3)},
	{"DM1CSR3",	64,	TX4927_DMAC_DMCSR(1, 3)},

	{"DM1MFDR",	64,	TX4927_DMAC_DMMFDR(1)},
	{"DM1MCR",	64,	TX4927_DMAC_DMMCR(1)},
	{NULL}
};

static tx4927_reg_proc_reg_t tx4927_reg_proc_pcic_regs[] =
{
	{"PCIID",	32,	TX4927_PCIC_PCIID},
	{"PCISTATUS",	32,	TX4927_PCIC_PCISTATUS},
	{"PCICCREV",	32,	TX4927_PCIC_PCICCREV},
	{"PCICFG1",	32,	TX4927_PCIC_PCICFG1},
	{"P2GM0PLBASE",	32,	TX4927_PCIC_P2GM0PLBASE},
	{"P2GM0PUBASE",	32,	TX4927_PCIC_P2GM0PUBASE},
	{"P2GM1PLBASE",	32,	TX4927_PCIC_P2GM1PLBASE},
	{"P2GM1PUBASE",	32,	TX4927_PCIC_P2GM1PUBASE},
	{"P2GM2BASE",	32,	TX4927_PCIC_P2GM2BASE},
	{"P2GIOPBASE",	32,	TX4927_PCIC_P2GIOPBASE},
	{"PCISID",	32,	TX4927_PCIC_PCISID},
	{"PCICAPPTR",	32,	TX4927_PCIC_PCICAPPTR},
	{"PCICFG2",	32,	TX4927_PCIC_PCICFG2},
	{"G2PTOCNT",	32,	TX4927_PCIC_G2PTOCNT},
	{"G2PSTATUS",	32,	TX4927_PCIC_G2PSTATUS},
	{"G2PMASK",	32,	TX4927_PCIC_G2PMASK},
	{"PCISSTAUS",	32,	TX4927_PCIC_PCISSTAUS},
	{"PCIMASK",	32,	TX4927_PCIC_PCIMASK},
	{"P2GCFG",	32,	TX4927_PCIC_P2GCFG},
	{"P2GSTATUS",	32,	TX4927_PCIC_P2GSTATUS},
	{"P2GMASK",	32,	TX4927_PCIC_P2GMASK},
	{"P2GCCMD",	32,	TX4927_PCIC_P2GCCMD},
	{"PBAREQPORT",	32,	TX4927_PCIC_PBAREQPORT},
	{"PBACFG",	32,	TX4927_PCIC_PBACFG},
	{"PBASTATUS",	32,	TX4927_PCIC_PBASTATUS},
	{"PBAMASK",	32,	TX4927_PCIC_PBAMASK},
	{"PBABM",	32,	TX4927_PCIC_PBABM},
	{"PBACREQ",	32,	TX4927_PCIC_PBACREQ},
	{"PBACGNT",	32,	TX4927_PCIC_PBACGNT},
	{"PBACSTATE",	32,	TX4927_PCIC_PBACSTATE},
	{"G2PM0GBASE",	64,	TX4927_PCIC_G2PM0GBASE},
	{"G2PM1GBASE",	64,	TX4927_PCIC_G2PM1GBASE},
	{"G2PM2GBASE",	64,	TX4927_PCIC_G2PM2GBASE},
	{"G2PIOGBASE",	64,	TX4927_PCIC_G2PIOGBASE},
	{"G2PM0MASK",	32,	TX4927_PCIC_G2PM0MASK},
	{"G2PM1MASK",	32,	TX4927_PCIC_G2PM1MASK},
	{"G2PM2MASK",	32,	TX4927_PCIC_G2PM2MASK},
	{"G2PIOMASK",	32,	TX4927_PCIC_G2PIOMASK},
	{"G2PM0PBASE",	64,	TX4927_PCIC_G2PM0PBASE},
	{"G2PM1PBASE",	64,	TX4927_PCIC_G2PM1PBASE},
	{"G2PM2PBASE",	64,	TX4927_PCIC_G2PM2PBASE},
	{"G2PIOPBASE",	64,	TX4927_PCIC_G2PIOPBASE},
	{"PCICCFG",	32,	TX4927_PCIC_PCICCFG},
	{"PCICSTATUS",	32,	TX4927_PCIC_PCICSTATUS},
	{"PCICMASK",	32,	TX4927_PCIC_PCICMASK},
	{"P2GM0GBASE",	64,	TX4927_PCIC_P2GM0GBASE},
	{"P2GM1GBASE",	64,	TX4927_PCIC_P2GM1GBASE},
	{"P2GM2GBASE",	64,	TX4927_PCIC_P2GM2GBASE},
	{"P2GIOGBASE",	64,	TX4927_PCIC_P2GIOGBASE},
	{"G2PCFGADRS",	32,	TX4927_PCIC_G2PCFGADRS},
	{"G2PCFGDATA",	32,	TX4927_PCIC_G2PCFGDATA},
	{"G2PINTACK",	32,	TX4927_PCIC_G2PINTACK},
/*	{"G2PSPC",	32,	TX4927_PCIC_G2PSPC}, */
	{"PCICDATA0",	32,	TX4927_PCIC_PCICDATA0},
	{"PCICDATA1",	32,	TX4927_PCIC_PCICDATA1},
	{"PCICDATA2",	32,	TX4927_PCIC_PCICDATA2},
	{"PCICDATA3",	32,	TX4927_PCIC_PCICDATA3},
	{"PDMCA",	64,	TX4927_PCIC_PDMCA},
	{"PDMGA",	64,	TX4927_PCIC_PDMGA},
	{"PDMPA",	64,	TX4927_PCIC_PDMPA},
	{"PDMCTR",	64,	TX4927_PCIC_PDMCTR},
	{"PDMCFG",	64,	TX4927_PCIC_PDMCFG},
	{"PDMSTATUS",	64,	TX4927_PCIC_PDMSTATUS},
	{NULL}
};

static tx4927_reg_proc_reg_t tx4927_reg_proc_config_regs[] =
{
	{"CCFG",	64,	TX4927_CONFIG_CCFG},
	{"REVID",	64,	TX4927_CONFIG_REVID},
	{"PCFG",	64,	TX4927_CONFIG_PCFG},
	{"TOEA",	64,	TX4927_CONFIG_TOEA},
	{"CLKCTR",	64,	TX4927_CONFIG_CLKCTR},
	{"GARBC",	64,	TX4927_CONFIG_GARBC},
	{"RAMP",	64,	TX4927_CONFIG_RAMP},
	{NULL}
};

static tx4927_reg_proc_reg_t tx4927_reg_proc_timer0_regs[] =
{
	{"TMTCR0",	32,	TX4927_TMR0_TMTCR0},
	{"TMTISR0",	32,	TX4927_TMR0_TMTISR0},
	{"TMCPRA0",	32,	TX4927_TMR0_TMCPRA0},
	{"TMCPRB0",	32,	TX4927_TMR0_TMCPRB0},
	{"TMITMR0",	32,	TX4927_TMR0_TMITMR0},
	{"TMCCDR0",	32,	TX4927_TMR0_TMCCDR0},
	{"TMPGMR0",	32,	TX4927_TMR0_TMPGMR0},
	{"TMTRR0",	32,	TX4927_TMR0_TMTRR0},
	{NULL}
};

static tx4927_reg_proc_reg_t tx4927_reg_proc_timer1_regs[] =
{
	{"TMTCR1",	32,	TX4927_TMR1_TMTCR1},
	{"TMTISR1",	32,	TX4927_TMR1_TMTISR1},
	{"TMCPRA1",	32,	TX4927_TMR1_TMCPRA1},
	{"TMCPRB1",	32,	TX4927_TMR1_TMCPRB1},
	{"TMITMR1",	32,	TX4927_TMR1_TMITMR1},
	{"TMCCDR1",	32,	TX4927_TMR1_TMCCDR1},
	{"TMPGMR1",	32,	TX4927_TMR1_TMPGMR1},
	{"TMTRR1",	32,	TX4927_TMR1_TMTRR1},
	{NULL}
};

static tx4927_reg_proc_reg_t tx4927_reg_proc_timer2_regs[] =
{
	{"TMTCR2",	32,	TX4927_TMR2_TMTCR2},
	{"TMTISR2",	32,	TX4927_TMR2_TMTISR2},
	{"TMCPRA2",	32,	TX4927_TMR2_TMCPRA2},
	{"TMCPRB2",	32,	TX4927_TMR2_TMCPRB2},
	{"TMITMR2",	32,	TX4927_TMR2_TMITMR2},
	{"TMCCDR2",	32,	TX4927_TMR2_TMCCDR2},
	{"TMPGMR2",	32,	TX4927_TMR2_TMPGMR2},
	{"TMTRR2",	32,	TX4927_TMR2_TMTRR2},
	{NULL}
};

static tx4927_reg_proc_reg_t tx4927_reg_proc_sio0_regs[] =
{
	{"SILCR0",	32,	TX4927_SIO0_SILCR0},
	{"SIDICR0",	32,	TX4927_SIO0_SIDICR0},
	{"SIDISR0",	32,	TX4927_SIO0_SIDISR0},
	{"SISCISR0",	32,	TX4927_SIO0_SISCISR0},
	{"SIFCR0",	32,	TX4927_SIO0_SIFCR0},
	{"SIFLCR0",	32,	TX4927_SIO0_SIFLCR0},
	{"SIBGR0",	32,	TX4927_SIO0_SIBGR0},
	{"SITFIFO0",	32,	TX4927_SIO0_SITFIF00},
	{"SIRFIFO0",	32,	TX4927_SIO0_SIRFIFO0},
	{NULL}
};

static tx4927_reg_proc_reg_t tx4927_reg_proc_sio1_regs[] =
{
	{"SILCR1",	32,	TX4927_SIO1_SILCR1},
	{"SIDICR1",	32,	TX4927_SIO1_SIDICR1},
	{"SIDISR1",	32,	TX4927_SIO1_SIDISR1},
	{"SISCISR1",	32,	TX4927_SIO1_SISCISR1},
	{"SIFCR1",	32,	TX4927_SIO1_SIFCR1},
	{"SIFLCR1",	32,	TX4927_SIO1_SIFLCR1},
	{"SIBGR1",	32,	TX4927_SIO1_SIBGR1},
	{"SITFIFO1",	32,	TX4927_SIO1_SITFIF01},
	{"SIRFIFO1",	32,	TX4927_SIO1_SIRFIFO1},
	{NULL}
};

static tx4927_reg_proc_reg_t tx4927_reg_proc_pio_regs[] =
{
	{"PIODO",	32,	TX4927_PIO_PIOD0},
	{"PIODI",	32,	TX4927_PIO_PIODI},
	{"PIODIR",	32,	TX4927_PIO_PIODIR},
	{"PIOOD",	32,	TX4927_PIO_PIOOD},
	{NULL}
};

static tx4927_reg_proc_reg_t tx4927_reg_proc_irc_regs[] =
{
	{"IRFLAG0",	32,	TX4927_IRC_IRFLAG0},
	{"IRFLAG1",	32,	TX4927_IRC_IRFLAG1},
	{"IRPOL",	32,	TX4927_IRC_IRPOL},
	{"IRRCNT",	32,	TX4927_IRC_IRRCNT},
	{"IRMASKINT",	32,	TX4927_IRC_IRMASKINT},
	{"IRMASKEXT",	32,	TX4927_IRC_IRMASKEXT},
	{"IRDEN",	32,	TX4927_IRC_IRDEN},
	{"IRDM0",	32,	TX4927_IRC_IRDM0},
	{"IRDM1",	32,	TX4927_IRC_IRDM1},
	{"IRLVL0",	32,	TX4927_IRC_IRLVL0},
	{"IRLVL1",	32,	TX4927_IRC_IRLVL1},
	{"IRLVL2",	32,	TX4927_IRC_IRLVL2},
	{"IRLVL3",	32,	TX4927_IRC_IRLVL3},
	{"IRLVL4",	32,	TX4927_IRC_IRLVL4},
	{"IRLVL5",	32,	TX4927_IRC_IRLVL5},
	{"IRLVL6",	32,	TX4927_IRC_IRLVL6},
	{"IRLVL7",	32,	TX4927_IRC_IRLVL7},
	{"IRMSK",	32,	TX4927_IRC_IRMSK},
	{"IREDC",	32,	TX4927_IRC_IREDC},
	{"IRPND",	32,	TX4927_IRC_IRPND},
	{"IRCS",	32,	TX4927_IRC_IRCS},
	{NULL}
};

static tx4927_reg_proc_reg_t tx4927_reg_proc_aclc_regs[] =
{
	{"ACCTLEN",	32,	TX4927_ACLC_ACCTLEN},
	{"ACCTLDIS",	32,	TX4927_ACLC_ACCTLDIS},
	{"ACREGACC",	32,	TX4927_ACLC_ACREGACC},
	{"ACINTSTS",	32,	TX4927_ACLC_ACINTSTS},
	{"ACINTMSTS",	32,	TX4927_ACLC_ACINTMSTS},
	{"ACINTEN",	32,	TX4927_ACLC_ACINTEN},
	{"ACINTDIS",	32,	TX4927_ACLC_ACINTDIS},
	{"ACSEMAPH",	32,	TX4927_ACLC_ACSEMAPH},
	{"ACGPIDAT",	32,	TX4927_ACLC_ACGPIDAT},
	{"ACGPODAT",	32,	TX4927_ACLC_ACGPODAT},
	{"ACSLTEN",	32,	TX4927_ACLC_ACSLTEN},
	{"ACSLTDIS",	32,	TX4927_ACLC_ACSLTDIS},
	{"ACFIFOSTS",	32,	TX4927_ACLC_ACFIFOSTS},
	{"ACDMASTS",	32,	TX4927_ACLC_ACDMASTS},
	{"ACDMASEL",	32,	TX4927_ACLC_ACDMASEL},
	{"ACAUDODAT",	32,	TX4927_ACLC_ACAUDODAT},
	{"ACSURRDAT",	32,	TX4927_ACLC_ACSURRDAT},
	{"ACCENTDAT",	32,	TX4927_ACLC_ACCENTDAT},
	{"ACLFEDAT",	32,	TX4927_ACLC_ACLFEDAT},
	{"ACAUDIDAT",	32,	TX4927_ACLC_ACAUDIDAT},
	{"ACMODODAT",	32,	TX4927_ACLC_ACMODODAT},
	{"ACMODIDAT",	32,	TX4927_ACLC_ACMODIDAT},
	{"ACREVID",	32,	TX4927_ACLC_ACREVID},
	{NULL}
};

static tx4927_reg_proc_module_t tx4927_reg_proc_modules[] =
{
	{"CONFIG",	tx4927_reg_proc_config_regs},
	{"SDRAMC",	tx4927_reg_proc_sdramc_regs},
	{"EBUSC",	tx4927_reg_proc_ebusc_regs},
	{"ECC",		tx4927_reg_proc_ecc_regs},
	{"DMAC0",	tx4927_reg_proc_dmac0_regs},
	{"DMAC1",	tx4927_reg_proc_dmac1_regs},
	{"PCIC",	tx4927_reg_proc_pcic_regs},
	{"TMR0",	tx4927_reg_proc_timer0_regs},
	{"TMR1",	tx4927_reg_proc_timer1_regs},
	{"TMR2",	tx4927_reg_proc_timer2_regs},
	{"SIO0",	tx4927_reg_proc_sio0_regs},
	{"SIO1",	tx4927_reg_proc_sio1_regs},
	{"PIO",		tx4927_reg_proc_pio_regs},
	{"IRC",		tx4927_reg_proc_irc_regs},
	{"ACLC",	tx4927_reg_proc_aclc_regs},
	{NULL}
};

/* ======================================================================= *
 * Inline function                                                         *
 * ======================================================================= */

/* ======================================================================= *
 * Linux driver interface                                                  *
 * ======================================================================= */

static struct file_operations tx4927_reg_proc_module_ops = {
	read:	tx4927_reg_proc_module_read,
};

static struct file_operations tx4927_reg_proc_reg_ops = {
	read:	tx4927_reg_proc_reg_read,
};

/* ------------------------ *
 * initialize/finalize      *
 * ------------------------ */

static int
tx4927_reg_proc_init(void)
{
	int res;
	int i, j = 0;
	tx4927_reg_proc_reg_t *regs;
	struct proc_dir_entry *dir;
	struct proc_dir_entry *ent;
	DPRFS;

	tx4927_reg_proc_root = proc_mkdir(TX4927_REG_PROC_DIRNAME, NULL);
	if (tx4927_reg_proc_root == NULL) {
		DPRFL(DPL_WARNING, "cannot create proc directory\n");
		res = -ENOMEM;
		goto err1;
	}

	for (i = 0; tx4927_reg_proc_modules[i].name != NULL; i++) {
		dir = proc_mkdir(tx4927_reg_proc_modules[i].name,
				 tx4927_reg_proc_root);
		if (dir == NULL) {
			DPRFL(DPL_WARNING, "cannot create module directory\n");
			res = -ENOMEM;
			goto err2;
		}
		tx4927_reg_proc_modules[i].dir = dir;
		ent = create_proc_entry(TX4927_REG_PROC_MODULE_FILENAME,
					S_IRUSR | S_IRGRP | S_IROTH, dir);
		if (ent == NULL) {
			DPRFL(DPL_WARNING,
			      "cannot create module entry\n");
			res = -ENOMEM;
			goto err3;
		}
		tx4927_reg_proc_modules[i].low_ino = ent->low_ino;
		ent->proc_fops = &tx4927_reg_proc_module_ops;

		regs = tx4927_reg_proc_modules[i].regs;
		for (j = 0; regs[j].name != NULL; j++) {
			ent = create_proc_entry(regs[j].name,
						S_IWUSR | S_IRUSR | S_IRGRP |
						S_IROTH, dir);
			if (ent == NULL) {
				DPRFL(DPL_WARNING,
				      "cannot create register entry\n");
				res = -ENOMEM;
				goto err4;
			}
			regs[j].low_ino = ent->low_ino;
			ent->proc_fops = &tx4927_reg_proc_reg_ops;
		}
	}

	DPRFE;
	return 0;

	do {
		regs = tx4927_reg_proc_modules[i].regs;
		do {
			remove_proc_entry(regs[j].name,
					  tx4927_reg_proc_modules[i].dir);
err4:
			j--;
		} while (j >= 0);
		remove_proc_entry(TX4927_REG_PROC_MODULE_FILENAME,
				  tx4927_reg_proc_modules[i].dir);
err3:
		remove_proc_entry(tx4927_reg_proc_modules[i].name,
				  tx4927_reg_proc_root);
err2:
		i--;
	} while (i >= 0);
	remove_proc_entry(TX4927_REG_PROC_DIRNAME, NULL);
err1:
	DPRFR;
	return res;
}

static void
tx4927_reg_proc_exit(void)
{
	int i, j;
	tx4927_reg_proc_reg_t *regs;
	DPRFS;

	for (i = 0; tx4927_reg_proc_modules[i].name != NULL; i++) {
		regs = tx4927_reg_proc_modules[i].regs;
		for (j = 0; regs[j].name != NULL; j++) {
			remove_proc_entry(regs[j].name,
					  tx4927_reg_proc_modules[i].dir);
		}
		remove_proc_entry(TX4927_REG_PROC_MODULE_FILENAME,
				  tx4927_reg_proc_modules[i].dir);
		remove_proc_entry(tx4927_reg_proc_modules[i].name,
				  tx4927_reg_proc_root);
	}
	remove_proc_entry(TX4927_REG_PROC_DIRNAME, NULL);

	DPRFE;
	return;
}

module_init(tx4927_reg_proc_init);
module_exit(tx4927_reg_proc_exit);

/* ------------------------ *
 * proc interface           *
 * ------------------------ */

static ssize_t
tx4927_reg_proc_module_read(struct file *filp, char *buf,
			    size_t count, loff_t *f_pos)
{
	int res;
	int i, j;
	tx4927_reg_proc_reg_t *regs;
	unsigned short low_ino;
	char tmp_buf[TX4927_REG_PROC_MAX_NAMELEN + 22];
	loff_t tmp_pos;
	ssize_t nwritten;
	int tmp_len;
	int ncopy;
	DPRFS;

	res = access_ok(VERIFY_WRITE, buf, count);
	if (!res) {
		DPRFR;
		return -EFAULT;
	}
	if (count == 0) {
		DPRFR;
		return 0;
	}

	low_ino = filp->f_dentry->d_inode->i_ino;
	for (i = 0; tx4927_reg_proc_modules[i].name != NULL; i++) {
		if (tx4927_reg_proc_modules[i].low_ino == low_ino) {
			break;
		}
	}
	if (tx4927_reg_proc_modules[i].name == NULL) {
		DPRFR;
		return -ENXIO;
	}
	regs = tx4927_reg_proc_modules[i].regs;

	nwritten = 0;
	tmp_pos = 0;
	for (j = 0; regs[j].name != NULL; j++) {
		switch (regs[j].bitwidth) {
		case 8:
			res = sprintf(tmp_buf, "%s\t=0x%02x\n", regs[j].name,
				      TX4927_RD08((unsigned long)
						  TX4927_REG(regs[j].offset)));
			break;
		case 16:
			res = sprintf(tmp_buf, "%s\t=0x%04x\n", regs[j].name,
				      TX4927_RD16((unsigned long)
						  TX4927_REG(regs[j].offset)));
			break;
		case 32:
			res = sprintf(tmp_buf, "%s\t=0x%08x\n", regs[j].name,
				      TX4927_RD32((unsigned long)
						  TX4927_REG(regs[j].offset)));
			break;
		case 64:
			res = sprintf(tmp_buf, "%s\t=0x%016llx\n",
				      regs[j].name,
				      TX4927_RD64((unsigned long)
						  TX4927_REG(regs[j].offset)));
			break;
		default:
			DPRFR;
			return -EINVAL;
		}
		tmp_len = res;
		if (tmp_pos == *f_pos) {
			ncopy = tmp_len;
			if (ncopy > count) {
				ncopy = count;
			}
			res = copy_to_user(buf + nwritten, tmp_buf, ncopy);
			if (res != 0) {
				DPRFR;
				return -EFAULT;
			}
			*f_pos += ncopy;
			nwritten += ncopy;
			count -= ncopy;
			if (count == 0) {
				break;
			}
		} else if (tmp_pos < *f_pos && tmp_pos + tmp_len > *f_pos) {
			ncopy = tmp_pos + tmp_len - *f_pos;
			if (ncopy > count) {
				ncopy = count;
			}
			res = copy_to_user(buf + nwritten,
					   tmp_buf + (*f_pos - tmp_pos),
					   ncopy);
			if (res != 0) {
				DPRFR;
				return -EFAULT;
			}
			*f_pos += ncopy;
			nwritten += ncopy;
			count -= ncopy;
			if (count == 0) {
				break;
			}
		}
		tmp_pos += tmp_len;
	}

	DPRFE;
	return nwritten;
}

static ssize_t
tx4927_reg_proc_reg_read(struct file *filp, char *buf,
			 size_t count, loff_t *f_pos)
{
	int res;
	int i, j;
	tx4927_reg_proc_reg_t *regs;
	DPRFS;
	unsigned short low_ino;
	char tmp_buf[TX4927_REG_PROC_MAX_NAMELEN + 22];
	ssize_t nwritten;
	int tmp_len;
	int ncopy;
	DPRFS;

	res = access_ok(VERIFY_WRITE, buf, count);
	if (!res) {
		DPRFR;
		return -EFAULT;
	}
	if (count == 0) {
		DPRFR;
		return 0;
	}

	low_ino = filp->f_dentry->d_inode->i_ino;
	for (i = 0; tx4927_reg_proc_modules[i].name != NULL; i++) {
		regs = tx4927_reg_proc_modules[i].regs;
		for (j = 0; regs[j].name != NULL; j++) {
			if (regs[j].low_ino == low_ino) {
				goto found;
			}
		}
	}
	DPRFR;
	return -ENXIO;

found:
	nwritten = 0;

	switch (regs[j].bitwidth) {
	case 8:
		res = sprintf(tmp_buf, "%s\t=0x%02x\n", regs[j].name,
			      TX4927_RD08((unsigned long)
					  TX4927_REG(regs[j].offset)));
		break;
	case 16:
		res = sprintf(tmp_buf, "%s\t=0x%04x\n", regs[j].name,
			      TX4927_RD16((unsigned long)
					  TX4927_REG(regs[j].offset)));
		break;
	case 32:
		res = sprintf(tmp_buf, "%s\t=0x%08x\n", regs[j].name,
			      TX4927_RD32((unsigned long)
					  TX4927_REG(regs[j].offset)));
		break;
	case 64:
		res = sprintf(tmp_buf, "%s\t=0x%016llx\n", regs[j].name,
			      TX4927_RD64((unsigned long)
					  TX4927_REG(regs[j].offset)));
		break;
	default:
		DPRFR;
		return -EINVAL;
	}
	tmp_len = res;
	if (*f_pos == 0) {
		ncopy = tmp_len;
		if (ncopy > count) {
			ncopy = count;
		}
		res = copy_to_user(buf, tmp_buf, ncopy);
		if (res != 0) {
			DPRFR;
			return -EFAULT;
		}
		*f_pos += ncopy;
		nwritten += ncopy;
		count -= ncopy;
	} else if (0 < *f_pos && tmp_len > *f_pos) {
		ncopy = tmp_len - *f_pos;
		if (ncopy > count) {
			ncopy = count;
		}
		res = copy_to_user(buf, tmp_buf + *f_pos, ncopy);
		if (res != 0) {
			DPRFR;
			return -EFAULT;
		}
		*f_pos += ncopy;
		nwritten += ncopy;
		count -= ncopy;
	}

	DPRFE;
	return nwritten;
}
