/* _FOR_OPTICAL_DISK_ */
/*
 *
 *  BRIEF MODULE DESCRIPTION
 *	Miscellaneous stuffs for Sony BSNC DIF-144 board. 
 * 
 *  Copyright 2002, 2003, 2004, 2005 Sony Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 */
/*
 *  Changes:
 *
 *  2002/12/16 Sony Corporation
 *  	* Added dif144_get_lighted_led().
 *  	* Added a argument for dif144_led_light() to specify lighting-mode such
 *  	  as OVERWRITE and OR.
 *  2004/12/07 Sony Corporation
 *  	* Restore registers of OSD.
 */

#include <asm/io.h>
#include <asm/vr41xx/dif144.h>
#include <asm/types.h>
#include <linux/module.h>
#include <linux/pci.h>

u8 led_cache = ~BSNC_DIF144_IO_DEFAULT & 0xff;
u16 tp_cache = (BSNC_DIF144_IO_DEFAULT & 0xff00) >> 8;
u8 dev_clock = BSNC_DIF144_IO_DEFAULT & 0xfc;

void dif144_led_light(u8 mask, u8 mode)
{
	u16 tp_and_led;

	switch (mode) {
	case BSNC_DIF144_LED_MODE_OVERWRITE:
		break;
	case BSNC_DIF144_LED_MODE_OR:
		mask |= led_cache;
		break;
	}

	led_cache = mask;

	/* 
	 * ON: 0, OFF: 1
	 */
#ifdef CONFIG_BSNC_CAM_PM
	/* bit15-8: TP, bit0-1: led, bit 2-7: clock */
	tp_and_led = (~mask & 0x3) | (dev_clock & 0xfc) | (tp_cache << 8);
#else
	/* bit15-8: TP, bit7-0: led */
	tp_and_led = (~mask & 0xff) | (tp_cache << 8);
#endif
        writew(tp_and_led, BSNC_DIF144_LED);
}


static u8 osd_latency_timer;
static u16 osd_command;
static u32 osd_base_addr;

static void
backup_osd_regs(struct pci_dev *dev)
{
	pci_read_config_byte(dev, PCI_LATENCY_TIMER, &osd_latency_timer);
	pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &osd_base_addr);
}

static void
restore_osd_regs(struct pci_dev *dev)
{
	u32 cmp32;
	u16 cmp16;
	u8 cmp8;
	int timeount = 5;

retry:
	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout(HZ);

	pci_write_config_dword(dev, PCI_BASE_ADDRESS_0, osd_base_addr);
	pci_write_config_word(dev, PCI_COMMAND, 0x7);
	pci_write_config_byte(dev, PCI_LATENCY_TIMER, osd_latency_timer);

	pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &cmp32);
	pci_read_config_byte(dev, PCI_LATENCY_TIMER, &cmp8);
	pci_read_config_word(dev, PCI_COMMAND, &cmp16);

	if (cmp32 == osd_base_addr && cmp16 == 0x7 && cmp8 == osd_latency_timer)
		return;

	timeount--;
	if (timeount == 0) {
		printk("Unable to restore the registers of OSD\n");
		return;
	}

	goto retry;
}

/* hi: supply clock */
void dif144_dev_clock(u8 dev, int on)
{
#ifdef CONFIG_BSNC_CAM_PM
	u16 val;
	struct pci_dev *p, *osd = NULL;
	signed long pon_wait_time ;
	
	if (dev == (1 << 5)) { /* OSD? */
		pci_for_each_dev(p) {
			if (p->vendor == PCI_VENDOR_ID_INTERG &&
			    p->device == PCI_DEVICE_ID_INTERG_5202) {
				osd = p;
			}
		}
		if (on) {
			dev_clock |= dev;
			val = (~led_cache & 0x03) | (dev_clock & 0xfc) | (tp_cache << 8);
			writew(val , BSNC_DIF144_LED);
			current->state = TASK_INTERRUPTIBLE;
			pon_wait_time = 1 + ( 2 * HZ ) / 100;
			while( pon_wait_time ){
				pon_wait_time = schedule_timeout( pon_wait_time );
			}
			dev_clock |= ( 1 << 3 );
		}
		else {
			dev |= ( 1 << 3 );
			dev_clock &= ~dev;
		}
	}
	else {
		if (on)
			dev_clock |= dev;
		else
			dev_clock &= ~dev;
	}
	
	val = (~led_cache & 0x03) | (dev_clock & 0xfc) | (tp_cache << 8);
	
	if (osd && !on) {
		backup_osd_regs(osd);
	}
	
	writew(val , BSNC_DIF144_LED);
	
	if (osd && on) {
		restore_osd_regs(osd);
	}
#endif
}

u8 dif144_get_lighted_led(void)
{
	return led_cache;
}

void dif144_tp_hi(u8 port)
{
	tp_cache |= 0x1 << port;
	writew((~led_cache & 0xff) | (tp_cache << 8), BSNC_DIF144_LED);
}

void dif144_tp_low(u8 port)
{
	tp_cache &= ~(0x1 << port);
	writew((~led_cache & 0xff) | (tp_cache << 8), BSNC_DIF144_LED);
}

u8 dif144_dipswitch(u16 mask)
{
        if (readl(BSNC_DIF144_DIPSW) & mask) {
                return BSNC_DIF144_DIPSW_OFF;
        } else {
                return BSNC_DIF144_DIPSW_ON;
        }
}

int dif144_boot_mode(void)
{
	if (readl(0xbffe4004) & 0x00020000) {
		return BSNC_DIF144_MODE_A;
	} else {
		return BSNC_DIF144_MODE_B;
	}
}

EXPORT_SYMBOL(dif144_dipswitch);

