/* _FOR_OPTICAL_DISK_ */

/*
 * $Id: dif144-flash.c,v 1.3 2004/12/15 04:58:56 utu_core Exp $
 *
 * Handle mapping of the flash on Sony BSNC DIF-144/DPR-238/DVP-27 boards
 *
 * Copyright 2002, 2003, 2004, 2005 Sony Corporation
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <asm/vr41xx/dif144.h>

#define ROM_ADDR_ROMCS0	KSEG1ADDR(0x1e000000)
#define ROM_ADDR_ROMCS2	KSEG1ADDR(0x1c000000)
#define ROM_SIZE	0x02000000	/* 32 MB */
#define BUSWIDTH	4		/* data bus size in octets */

/* Flash / Partition sizing */
#define FS_PART_A_SIZE_KiB        17920 /* BA   0-160 0xbe000000-0xbf17ffff */
#define FS_PART_B_SIZE_KiB        10752 /* BA 161-244 0xbf180000-0xbfbfffff */
#define FS_PART_LOADER_SIZE_KiB     128 /* BA     245 0xbfc00000-0xbfc1ffff */ 
#define FS_PART_KERNEL_A_SIZE_KiB  2048 /* BA 246-283 0xbfc20000-0xbfffffff */ 
#define FS_PART_KERNEL_B_SIZE_KiB  1792 /* BA 246-283 0xbfc20000-0xbfffffff */ 
#define FS_PART_MAC_ADDR_SIZE_KiB   128 /* BA 246-283 0xbfc20000-0xbfffffff */ 
/* partition_info gives details on the logical partitions that the split the
 * single flash device into. If the size if zero we use up to the end of the
 * device. */
#define NUM_PARTITIONS 6

static struct mtd_partition partition_info[]=
{
	{ 
		name: "File system partition A",
		offset: 0,
		size: FS_PART_A_SIZE_KiB * 1024
	}, {
		name: "File system partition B",
		offset: FS_PART_A_SIZE_KiB * 1024,
		size:   FS_PART_B_SIZE_KiB * 1024
	}, {
		name: "Boot loader partition",
		offset: (FS_PART_A_SIZE_KiB +
			 FS_PART_B_SIZE_KiB) * 1024,
		size:    FS_PART_LOADER_SIZE_KiB * 1024
	}, {
		name: "Kernel partition A",
		offset: (FS_PART_A_SIZE_KiB +
			 FS_PART_B_SIZE_KiB +
			 FS_PART_LOADER_SIZE_KiB) * 1024,
		size:    FS_PART_KERNEL_A_SIZE_KiB * 1024
	}, {
		name: "Kernel partition B",
		offset: (FS_PART_A_SIZE_KiB +
			 FS_PART_B_SIZE_KiB +
			 FS_PART_LOADER_SIZE_KiB +
			 FS_PART_KERNEL_A_SIZE_KiB) * 1024,
		size: FS_PART_KERNEL_B_SIZE_KiB * 1024
	}, {
		name: "Ethernet MAC address",
		offset: (FS_PART_A_SIZE_KiB +
			 FS_PART_B_SIZE_KiB +
			 FS_PART_LOADER_SIZE_KiB +
			 FS_PART_KERNEL_A_SIZE_KiB +
			 FS_PART_KERNEL_B_SIZE_KiB) * 1024
	} 
};
static struct mtd_info *mtd_info_onboard;
static struct mtd_info *mtd_info_jig;

__u8 dif144_read8(struct map_info *map, unsigned long ofs)
{
	return readb(map->map_priv_1 + ofs);
}

__u16 dif144_read16(struct map_info *map, unsigned long ofs)
{
	return readw(map->map_priv_1 + ofs);
}

__u32 dif144_read32(struct map_info *map, unsigned long ofs)
{
	return readl(map->map_priv_1 + ofs);
}

void dif144_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, (void *)(map->map_priv_1 + from), len);
}

void dif144_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	writeb(d, map->map_priv_1 + adr);
}

void dif144_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	writew(d, map->map_priv_1 + adr);
}

void dif144_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	writel(d, map->map_priv_1 + adr);
}

void dif144_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio((void *)(map->map_priv_1 + to), from, len);
}

struct map_info dif144_map_onboard = {
	name: "DIF-144/DPR-238/DVP-27 Flash-ROM on board",
	size: ROM_SIZE,
	buswidth: BUSWIDTH,
	read8: dif144_read8,
	read16: dif144_read16,
	read32: dif144_read32,
	copy_from: dif144_copy_from,
	write8: dif144_write8,
	write16: dif144_write16,
	write32: dif144_write32,
	copy_to: dif144_copy_to
};

struct map_info dif144_map_jigrom = {
	name: "DIF-144 DIF-144/DPR-238/DVP-27 JIG-ROM",
	size: ROM_SIZE,
	buswidth: BUSWIDTH,
	read8: dif144_read8,
	read16: dif144_read16,
	read32: dif144_read32,
	copy_from: dif144_copy_from,
	write8: dif144_write8,
	write16: dif144_write16,
	write32: dif144_write32,
	copy_to: dif144_copy_to
};

static struct mtd_info *do_map (struct map_info *map, unsigned long start)
{
	struct mtd_info *mymtd;

	printk(KERN_NOTICE
	       "Sony BSNC DIF-144/DPR-238/DVP-27 flash device: %x at %x\n",
	       map->size, start);

	map->map_priv_1 = start;
	mymtd = do_map_probe("jedec_probe", map);
	if (mymtd) {
		mymtd->module = THIS_MODULE;
		add_mtd_device(mymtd);
		add_mtd_partitions(mymtd, partition_info, NUM_PARTITIONS);
	}

	return mymtd;
}


int __init init_dif144_flash(void)
{
	int jigboot, jigexist;

	jigexist = dif144_dipswitch(BSNC_DIF144_PIN_JIG_EXIST) ? 1 : 0;
#ifndef CONFIG_MIPS_BSNC_DIF144_JIG_BOOT_OLD
	jigboot = dif144_dipswitch(BSNC_DIF144_DIPSW_JIG_BOOT) ? 1 : 0;
#else
	jigboot = dif144_dipswitch(BSNC_DIF144_DIPSW_JIG_BOOT_OLD) ? 1 : 0;
#endif

	mtd_info_onboard = do_map(&dif144_map_onboard,
				  jigboot ? ROM_ADDR_ROMCS2 : ROM_ADDR_ROMCS0);
	if (!mtd_info_onboard) {
		return -ENXIO;
	}

	if (jigexist) {
		mtd_info_jig = do_map(&dif144_map_jigrom,
				      jigboot ? ROM_ADDR_ROMCS0 : ROM_ADDR_ROMCS2);
		if (!mtd_info_jig) {
			return -ENXIO;
		}
	}
		
	return 0;
}

static void __exit cleanup_dif144_flash(void)
{
	if (mtd_info_onboard) {
		del_mtd_device(mtd_info_onboard);
		map_destroy(mtd_info_onboard);
	}
	if (mtd_info_jig) {
		del_mtd_device(mtd_info_jig);
		map_destroy(mtd_info_jig);
	}
	dif144_map_onboard.map_priv_1 = 0;
	dif144_map_jigrom.map_priv_1 = 0;
}

module_init(init_dif144_flash);
module_exit(cleanup_dif144_flash);

MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION("MTD map driver for Sony BSNC DIF-144/DPR-238/DVP-27 board");
MODULE_LICENSE("GPL");

