/*
 * Copyright 2003 Sony Corporation.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "ecc.h"

/* ======================================================================= *
 * constant (internal)                                                     *
 * ======================================================================= */

#ifndef NULL
#define NULL	0
#endif

#if !defined(ECC_CALC_SIMPLE) && !defined(ECC_CALC_FAST)
#define ECC_CALC_FAST	1
#endif

/* ======================================================================= *
 * Debug print function from dpr package                                   *
 * ======================================================================= */

#define DPL_EMERG	0
#define DPL_ALERT	1
#define DPL_CRIT	2
#define DPL_ERR		3
#define DPL_WARNING	4
#define DPL_NOTICE	5
#define DPL_INFO	6
#define DPL_DEBUG	7

#define DPM_PORPOSE	0x0000000f
#define DPM_ERR		0x00000001
#define DPM_TRACE	0x00000002
#define DPM_LAYER	0x000000f0
#define DPM_APPLICATION	0x00000010
#define DPM_MIDDLEWARE	0x00000020
#define DPM_DRIVER	0x00000040
#define DPM_KERNEL	0x00000080
#define DPM_USERDEF	0xffff0000

#ifdef DEBUG

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DPL_DEBUG
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  DPM_PORPOSE
#endif

#ifndef DPR
#include <stdio.h>
#define DPR(fmt, args...)	fprintf(stderr, fmt, ## args)
#endif

#define DPRL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPR(fmt, ## args))
#define DPRM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPR(fmt, ## args))
#define DPRLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPR(fmt, ## args))
#define DPRC(cond, fmt, args...) \
	(void)((cond) && DPR(fmt, ## args))
#define DPRF(fmt, args...)	DPR(__FUNCTION__ ": " fmt, ## args)
#define DPRFL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPRF(fmt, ## args))
#define DPRFM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPRF(fmt, ## args))
#define DPRFLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPRF(fmt, ## args))
#define DPRFC(cond, fmt, args...) \
	(void)((cond) && DPRF(fmt, ## args))

#define	DPRFS	DPRFLM(DPL_DEBUG, DPM_TRACE, "start\n")
#define	DPRFE	DPRFLM(DPL_DEBUG, DPM_TRACE, "end\n")
#define DPRFR	DPRFLM(DPL_DEBUG, DPM_TRACE, "return\n")

#else  /* DEBUG */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL -1
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  0
#endif

#ifndef DPR
#define DPR(fmt, args...)
#endif

#define DPRL(level, fmt, args...)
#define DPRM(mask, fmt, args...)
#define DPRLM(level, mask, fmt, args...)
#define DPRC(cond, fmt, args...)
#define DPRF(fmt, args...)
#define DPRFL(level, fmt, args...)
#define DPRFM(mask, fmt, args...)
#define DPRFLM(level, mask, fmt, args...)
#define DPRFC(cond, fmt, args...)

#define	DPRFS
#define	DPRFE
#define DPRFR

#endif /* DEBUG */

/* ======================================================================= *
 * ECC library interface                                                   *
 * ======================================================================= */

#if defined(ECC_CALC_SIMPLE) || defined(ECC_CALC_BOTH)
int
#ifdef ECC_CALC_BOTH
ecc_calc_simple(const void *data, unsigned int size, unsigned long *ecc)
#else
ecc_calc(const void *data, unsigned int size, unsigned long *ecc)
#endif
{
	int i, j;
	unsigned long ecc_tmp;
	unsigned long bi_cnt;
	unsigned char c;
	unsigned long mask;
	DPRFS;

	if (size > ECC_DATASIZE_MAX) {
		DPRFR;
		return -EECCINVAL;
	}

	ecc_tmp = 0;
	bi_cnt = 0;
	for (i = 0; i < size; i++) {
		c = *((const char *)data + i);
		for (j = 0; j < 8; j++) {
			if (c & 1) {
				ecc_tmp ^= (0x55555555 + bi_cnt);
			}
			c >>= 1;
			bi_cnt = ((bi_cnt * 3 + 1) & 0x55555555);
		}
	}

	for (i = 1, mask = 0xffffffc0; i < size; i <<= 1, mask <<= 2);
	*ecc = (ecc_tmp & ~mask);

	DPRFE;
	return 0;
}
#endif /* ECC_CALC_SIMPLE || ECC_CALC_BOTH */

#ifdef ECC_CALC_FAST
static signed char
ecc_byte_table[256] = {
	0x00, 0x95, 0x96, 0x03, 0x99, 0x0c, 0x0f, 0x9a,
	0x9a, 0x0f, 0x0c, 0x99, 0x03, 0x96, 0x95, 0x00,
	0xa5, 0x30, 0x33, 0xa6, 0x3c, 0xa9, 0xaa, 0x3f,
	0x3f, 0xaa, 0xa9, 0x3c, 0xa6, 0x33, 0x30, 0xa5,
	0xa6, 0x33, 0x30, 0xa5, 0x3f, 0xaa, 0xa9, 0x3c,
	0x3c, 0xa9, 0xaa, 0x3f, 0xa5, 0x30, 0x33, 0xa6,
	0x03, 0x96, 0x95, 0x00, 0x9a, 0x0f, 0x0c, 0x99,
	0x99, 0x0c, 0x0f, 0x9a, 0x00, 0x95, 0x96, 0x03,
	0xa9, 0x3c, 0x3f, 0xaa, 0x30, 0xa5, 0xa6, 0x33,
	0x33, 0xa6, 0xa5, 0x30, 0xaa, 0x3f, 0x3c, 0xa9,
	0x0c, 0x99, 0x9a, 0x0f, 0x95, 0x00, 0x03, 0x96,
	0x96, 0x03, 0x00, 0x95, 0x0f, 0x9a, 0x99, 0x0c,
	0x0f, 0x9a, 0x99, 0x0c, 0x96, 0x03, 0x00, 0x95,
	0x95, 0x00, 0x03, 0x96, 0x0c, 0x99, 0x9a, 0x0f,
	0xaa, 0x3f, 0x3c, 0xa9, 0x33, 0xa6, 0xa5, 0x30,
	0x30, 0xa5, 0xa6, 0x33, 0xa9, 0x3c, 0x3f, 0xaa,
	0xaa, 0x3f, 0x3c, 0xa9, 0x33, 0xa6, 0xa5, 0x30,
	0x30, 0xa5, 0xa6, 0x33, 0xa9, 0x3c, 0x3f, 0xaa,
	0x0f, 0x9a, 0x99, 0x0c, 0x96, 0x03, 0x00, 0x95,
	0x95, 0x00, 0x03, 0x96, 0x0c, 0x99, 0x9a, 0x0f,
	0x0c, 0x99, 0x9a, 0x0f, 0x95, 0x00, 0x03, 0x96,
	0x96, 0x03, 0x00, 0x95, 0x0f, 0x9a, 0x99, 0x0c,
	0xa9, 0x3c, 0x3f, 0xaa, 0x30, 0xa5, 0xa6, 0x33,
	0x33, 0xa6, 0xa5, 0x30, 0xaa, 0x3f, 0x3c, 0xa9,
	0x03, 0x96, 0x95, 0x00, 0x9a, 0x0f, 0x0c, 0x99,
	0x99, 0x0c, 0x0f, 0x9a, 0x00, 0x95, 0x96, 0x03,
	0xa6, 0x33, 0x30, 0xa5, 0x3f, 0xaa, 0xa9, 0x3c,
	0x3c, 0xa9, 0xaa, 0x3f, 0xa5, 0x30, 0x33, 0xa6,
	0xa5, 0x30, 0x33, 0xa6, 0x3c, 0xa9, 0xaa, 0x3f,
	0x3f, 0xaa, 0xa9, 0x3c, 0xa6, 0x33, 0x30, 0xa5,
	0x00, 0x95, 0x96, 0x03, 0x99, 0x0c, 0x0f, 0x9a,
	0x9a, 0x0f, 0x0c, 0x99, 0x03, 0x96, 0x95, 0x00,
};

int
ecc_calc(const void *data, unsigned int size, unsigned long *ecc)
{
	int i;
	unsigned long ecc_tmp;
	unsigned long bi_cnt;
	unsigned char c;
	unsigned long mask;
	signed int ecc_byte;
	DPRFS;

	if (size > ECC_DATASIZE_MAX) {
		DPRFR;
		return -EECCINVAL;
	}

	ecc_tmp = 0;
	bi_cnt = 0;
	for (i = 0; i < size; i++) {
		c = *((const char *)data + i);
		ecc_byte = ecc_byte_table[c];
		ecc_tmp ^= (ecc_byte & 0x3f);
		if (ecc_byte < 0) {
			ecc_tmp ^= (0x55555540 + bi_cnt);
		}
		bi_cnt = ((bi_cnt * 3 + 0x40) & 0x55555540);
	}

	for (i = 1, mask = 0xffffffc0; i < size; i <<= 1, mask <<= 2);
	*ecc = (ecc_tmp & ~mask);

	DPRFE;
	return 0;
}
#endif /* ECC_CALC_FAST */

int
ecc_correct(const void *data, unsigned int size,
	    unsigned long ecc1, unsigned long ecc2)
{
	unsigned long ecc_remain;
	unsigned long error_bit;
	int ecc_pair_xor;
	int nerrors;
	int i;
	DPRFS;

	if (size > ECC_DATASIZE_MAX) {
		DPRFR;
		return -EECCINVAL;
	}
	if (size == 0) {
		DPRFR;
		return 0;
	}
	if (ecc1 == ecc2) {
		DPRFR;
		return 0;
	}

	error_bit = 0;
	nerrors = 0;
	ecc_remain = size * 8 - 1;
	for (i = 0;; i++) {
		if (ecc_remain == 0) {
			break;
		}
		ecc_pair_xor = ((ecc1 & 0x3) ^ (ecc2 & 0x3));
		switch (ecc_pair_xor) {
		case 0:
			break;
		case 1:
		case 2:
			error_bit |= ((ecc_pair_xor >> 1) << i);
			nerrors++;
			break;
		case 3:
			DPRFR;
			return -EECCCORRECT;
		}
		ecc1 >>= 2;
		ecc2 >>= 2;
		ecc_remain >>= 1;
	}
	if (nerrors == i) {
		*((unsigned char *)data + error_bit / 8) ^=
			(1 << (error_bit & 0x7));
		DPRFR;
		return 1;
	} else if (nerrors == 1) {
		DPRFR;
		return 0;
	} else if (nerrors > 0) {
		DPRFR;
		return -EECCCORRECT;
	}

	DPRFE;
	return 0;
}
