/*
 *  pci_conf.c -- do PCI configuration cycle
 *
 *  Copyright 2003 Sony Corporation.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/pci_ids.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/pci_conf.h>

#define DEBUG		1
#define DEBUG_LEVEL	DPL_NOTICE

#ifdef DEBUG
#define DPR(fmt, args...)	printk(fmt, ## args)
#endif /* DEBUG */

/* ======================================================================= *
 * Debug print function from dpr package                                   *
 * ======================================================================= */

#define DPL_EMERG	0
#define DPL_ALERT	1
#define DPL_CRIT	2
#define DPL_ERR		3
#define DPL_WARNING	4
#define DPL_NOTICE	5
#define DPL_INFO	6
#define DPL_DEBUG	7

#define DPM_PORPOSE	0x0000000f
#define DPM_ERR		0x00000001
#define DPM_TRACE	0x00000002
#define DPM_LAYER	0x000000f0
#define DPM_APPLICATION	0x00000010
#define DPM_MIDDLEWARE	0x00000020
#define DPM_DRIVER	0x00000040
#define DPM_KERNEL	0x00000080
#define DPM_USERDEF	0xffff0000

#ifdef DEBUG

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL DPL_DEBUG
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  DPM_PORPOSE
#endif

#ifndef DPR
#include <stdio.h>
#define DPR(fmt, args...)	fprintf(stderr, fmt, ## args)
#endif

#define DPRL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPR(fmt, ## args))
#define DPRM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPR(fmt, ## args))
#define DPRLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPR(fmt, ## args))
#define DPRC(cond, fmt, args...) \
	(void)((cond) && DPR(fmt, ## args))
#define DPRF(fmt, args...)	DPR("%s: " fmt, __FUNCTION__, ## args)
#define DPRFL(level, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && DPRF(fmt, ## args))
#define DPRFM(mask, fmt, args...) \
	(void)(((mask) & DEBUG_MASK) && DPRF(fmt, ## args))
#define DPRFLM(level, mask, fmt, args...) \
	(void)((level) <= DEBUG_LEVEL && ((mask) & DEBUG_MASK) && \
		DPRF(fmt, ## args))
#define DPRFC(cond, fmt, args...) \
	(void)((cond) && DPRF(fmt, ## args))

#define	DPRFS	DPRFLM(DPL_DEBUG, DPM_TRACE, "start\n")
#define	DPRFE	DPRFLM(DPL_DEBUG, DPM_TRACE, "end\n")
#define DPRFR	DPRFLM(DPL_DEBUG, DPM_TRACE, "return\n")

#else  /* DEBUG */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL -1
#endif
#ifndef DEBUG_MASK
#define DEBUG_MASK  0
#endif

#ifndef DPR
#define DPR(fmt, args...)
#endif

#define DPRL(level, fmt, args...)
#define DPRM(mask, fmt, args...)
#define DPRLM(level, mask, fmt, args...)
#define DPRC(cond, fmt, args...)
#define DPRF(fmt, args...)
#define DPRFL(level, fmt, args...)
#define DPRFM(mask, fmt, args...)
#define DPRFLM(level, mask, fmt, args...)
#define DPRFC(cond, fmt, args...)

#define	DPRFS
#define	DPRFE
#define DPRFR

#endif /* DEBUG */

/* ======================================================================= *
 * Prototype declaration                                                   *
 * ======================================================================= */

static u32 pciconf_align(u32 val, u32 align);
static u32 pciconf_align_pow2(u32 val);

static struct resource *pciconf_malloc_res(const char *name,
					   unsigned long flags);

static int pciconf_allocate_resource(struct resource *resources[],
				     struct resource *new,
				     unsigned long size, unsigned long align);

static void pciconf_scan_bus(int *busnum, struct pci_bus *parent_bus,
			     struct pci_ops *ops,
			     u32 *io_size, u32 *mem_size, u32 *pmem_size);
static int pciconf_alloc_tree(int busnum, struct pci_bus *parent_bus,
			      struct pci_ops *ops,
			      struct resource *io_resources[],
			      struct resource *mem_resources[],
			      struct resource *pmem_resources[]);

static void pciconf_add_barsize(struct pci_dev *dev, struct pci_ops *ops,
				size_t num,
				u32 *io_size, u32 *mem_size, u32 *pmem_size);
static int pciconf_alloc_bar(struct pci_dev *dev, struct pci_ops *ops,
			     size_t num,
			     struct resource *io_resources[],
			     struct resource *mem_resources[],
			     struct resource *pmem_resources[]);

static void pciconf_preset_bridge_win(struct pci_dev *dev, struct pci_ops *ops,
				      u32 io_size, u32 mem_size,
				      u32 pmem_size);
static int pciconf_alloc_bridge_win(struct pci_dev *dev, struct pci_ops *ops,
				    struct resource *io_resources[],
				    struct resource *mem_resources[],
				    struct resource *pmem_resources[]);

static void pciconf_preset_cb_win(struct pci_dev *dev, struct pci_ops *ops,
				  u32 io_size, u32 mem_size, u32 pmem_size);

/* ======================================================================= *
 * Kernel service routine                                                  *
 * ======================================================================= */

int
pciconf_conf(int busnum, struct pci_ops *ops,
	     struct resource *io_resources[], struct resource *mem_resources[])
{
	int res;
	int busnum_max;
	u32 io_size;
	u32 mem_size;
	u32 pmem_size;
	DPRFS;

	busnum_max = busnum;
	pciconf_scan_bus(&busnum_max, NULL, ops,
			 &io_size, &mem_size, &pmem_size);

	res = pciconf_alloc_tree(busnum, NULL, ops,
				 io_resources, mem_resources, mem_resources);
	if (res < 0) {
		DPRFR;
		return res;
	}

	DPRFE;
	return (busnum_max + 1);
}

/* ======================================================================= *
 * Internal function                                                       *
 * ======================================================================= */

static u32
pciconf_align(u32 val, u32 align)
{
	return ((val + (align - 1)) & ~(align - 1));
}

static u32
pciconf_align_pow2(u32 val)
{
	int i;

	for (i = 0; i < 32; i++) {
		if (val <= ((u32)1 << i)) {
			break;
		}
	}

	return (1 << i);
}

static struct resource *
pciconf_malloc_res(const char *name, unsigned long flags)
{
	struct resource *res;
	DPRFS;

	res = kmalloc(sizeof(struct resource), GFP_KERNEL);
	if (res == NULL) {
		DPRFR;
		return NULL;
	}
	memset(res, 0, sizeof(struct resource));
	res->name = name;
	res->flags = flags;

	DPRFE;
	return res;
}

static int
pciconf_allocate_resource(struct resource *resources[], struct resource *new,
			  unsigned long size, unsigned long align)
{
	int res;
	int i;
	DPRFS;

	for (i = 0;; i++) {
		if (resources[i] == NULL) {
			res = -EBUSY;
			goto err;
		}
		res = allocate_resource(resources[i], new, size,
					0, -1, align, NULL, NULL);
		if (res == -EBUSY) {
			continue;
		} else if (res < 0) {
			goto err;
		}
		break;
	}

	DPRFE;
	return 0;

err:
	DPRFR;
	return res;
}

static void
pciconf_scan_bus(int *busnum, struct pci_bus *parent_bus,
		 struct pci_ops *ops,
		 u32 *io_size, u32 *mem_size, u32 *pmem_size)
{
	int res;
	struct pci_bus bus;
	struct pci_dev dev;
	u16 vendor_id;
	u8  header_type;
	u32 class_revision;
	u16 class_high;
	DPRFS;

	bus.parent = parent_bus;
	bus.number = *busnum;

	bus.ops = ops;
	dev.bus = &bus;

	for (dev.devfn = 0; dev.devfn < 0x100; dev.devfn++) {
		res = ops->read_word(&dev, PCI_VENDOR_ID, &vendor_id);
		if (res == -EBUSY) {
			continue;
		} else if (res != 0 || vendor_id == 0xffff) {
			if (PCI_FUNC(dev.devfn) == 0) {
				dev.devfn += 7;
			}
			continue;
		}
		ops->read_byte(&dev, PCI_HEADER_TYPE, &header_type);
		ops->read_dword(&dev, PCI_CLASS_REVISION, &class_revision);
		class_high = (class_revision >> 16);

		if ((header_type & 0x7f) == PCI_HEADER_TYPE_NORMAL &&
		    class_high != PCI_CLASS_BRIDGE_PCI) {
			/* normal device function */
#ifdef CONFIG_PCI_CONF_SKIP_LEGACY_IDE
			if (class_high == PCI_CLASS_STORAGE_IDE &&
			    (class_revision & 0x00000500) == 0) {
				DPRFL(DPL_DEBUG, "skip [%02x:%02x.%x]\n",
				      dev.bus->number, PCI_SLOT(dev.devfn),
				      PCI_FUNC(dev.devfn));
				goto break_next;
			}
#endif
			pciconf_add_barsize(&dev, ops, 6,
					    io_size, mem_size, pmem_size);
		} else if ((header_type & 0x7f) == PCI_HEADER_TYPE_BRIDGE &&
			   class_high == PCI_CLASS_BRIDGE_PCI) {
			/* PCI-to-PCI bridge */
			u32 sub_io_size = 0;
			u32 sub_mem_size = 0;
			u32 sub_pmem_size = 0;

			pciconf_add_barsize(&dev, ops, 2,
					    io_size, mem_size, pmem_size);

			ops->write_byte(&dev, PCI_PRIMARY_BUS, *busnum);
			(*busnum)++;
			ops->write_byte(&dev, PCI_SECONDARY_BUS, *busnum);
			ops->write_byte(&dev, PCI_SUBORDINATE_BUS, 0xff);

			pciconf_scan_bus(busnum, &bus, ops,
					 &sub_io_size, &sub_mem_size,
					 &sub_pmem_size);
			sub_io_size = pciconf_align(sub_io_size, 0x1000);
			sub_mem_size = pciconf_align(sub_mem_size, 0x100000);
			sub_pmem_size = pciconf_align(sub_pmem_size, 0x100000);

			ops->write_byte(&dev, PCI_SUBORDINATE_BUS, *busnum);
			pciconf_preset_bridge_win(&dev, ops, sub_io_size,
						  sub_mem_size, sub_pmem_size);

			*io_size += pciconf_align_pow2(sub_io_size);
			*mem_size += pciconf_align_pow2(sub_mem_size);
			*pmem_size += pciconf_align_pow2(sub_pmem_size);
		} else if ((header_type & 0x7f) == PCI_HEADER_TYPE_CARDBUS &&
			   class_high == PCI_CLASS_BRIDGE_CARDBUS) {
			/* CardBus bridge */
			pciconf_add_barsize(&dev, ops, 1,
					    io_size, mem_size, pmem_size);

			ops->write_byte(&dev, PCI_CB_PRIMARY_BUS, *busnum);
			(*busnum)++;
			ops->write_byte(&dev, PCI_CB_CARD_BUS, *busnum);
			ops->write_byte(&dev, PCI_CB_SUBORDINATE_BUS, *busnum);
			pciconf_preset_cb_win(&dev, ops, 0, 0, 0);
		} else {
			/* unknown device function */
			DPRFL(DPL_NOTICE,
			      "PCI[%02x:%02x.%x] Unknown PCI header type.\n",
			      dev.bus->number, PCI_SLOT(dev.devfn),
			      PCI_FUNC(dev.devfn));
		}
	break_next:
		if (!(header_type & 0x80) && PCI_FUNC(dev.devfn) == 0) {
			dev.devfn += 7;
		}
	}

	DPRFE;
	return;
}

static int
pciconf_alloc_tree(int busnum, struct pci_bus *parent_bus, struct pci_ops *ops,
		   struct resource *io_resources[],
		   struct resource *mem_resources[],
		   struct resource *pmem_resources[])
{
	int res;
	struct pci_bus bus;
	struct pci_dev dev;
	u16 vendor_id;
	u8  header_type;
	u32 class_revision;
	u16 class_high;
	DPRFS;

	bus.parent = parent_bus;
	bus.number = busnum;

	bus.ops = ops;
	dev.bus = &bus;

	for (dev.devfn = 0; dev.devfn < 0x100; dev.devfn++) {
		res = ops->read_word(&dev, PCI_VENDOR_ID, &vendor_id);
		if (res == -EBUSY) {
			continue;
		} else if (res != 0 || vendor_id == 0xffff) {
			if (PCI_FUNC(dev.devfn) == 0) {
				dev.devfn += 7;
			}
			continue;
		}
		ops->read_byte(&dev, PCI_HEADER_TYPE, &header_type);
		ops->read_dword(&dev, PCI_CLASS_REVISION, &class_revision);
		class_high = (class_revision >> 16);

		if ((header_type & 0x7f) == PCI_HEADER_TYPE_NORMAL &&
		    class_high != PCI_CLASS_BRIDGE_PCI) {
			/* normal device function */
#ifdef CONFIG_PCI_CONF_SKIP_LEGACY_IDE
			if (class_high == PCI_CLASS_STORAGE_IDE &&
			    (class_revision & 0x00000500) == 0) {
				DPRFL(DPL_DEBUG, "skip [%02x:%02x.%x]\n",
				      dev.bus->number, PCI_SLOT(dev.devfn),
				      PCI_FUNC(dev.devfn));
				goto break_next;
			}
#endif
			res = pciconf_alloc_bar(&dev, ops, 6, io_resources,
						mem_resources, pmem_resources);
			if (res < 0) {
				goto err;
			}
		} else if ((header_type & 0x7f) == PCI_HEADER_TYPE_BRIDGE &&
			   class_high == PCI_CLASS_BRIDGE_PCI) {
			/* PCI-to-PCI bridge */
			res = pciconf_alloc_bar(&dev, ops, 2, io_resources,
						mem_resources, pmem_resources);
			if (res < 0) {
				goto err;
			}
			res = pciconf_alloc_bridge_win(&dev, ops,
						       io_resources,
						       mem_resources,
						       pmem_resources);
			if (res < 0) {
				goto err;
			}
		} else if ((header_type & 0x7f) == PCI_HEADER_TYPE_CARDBUS &&
			   class_high == PCI_CLASS_BRIDGE_CARDBUS) {
			/* CardBus bridge */
			res = pciconf_alloc_bar(&dev, ops, 1, io_resources,
						mem_resources, pmem_resources);
			if (res < 0) {
				goto err;
			}
		} else {
			/* unknown device function */
			/**** do nothing ****/
		}
	break_next:
		if (!(header_type & 0x80) && PCI_FUNC(dev.devfn) == 0) {
			dev.devfn += 7;
		}
	}

	DPRFE;
	return 0;

err:
	DPRFR;
	return res;
}

static void
pciconf_add_barsize(struct pci_dev *dev, struct pci_ops *ops, size_t num,
		    u32 *io_size, u32 *mem_size, u32 *pmem_size)
{
	int i;
	u32 bar;
	int is_64bits_memory;
	DPRFS;

	for (i = 0; i < num; i++) {
		is_64bits_memory = 0;

		ops->write_dword(dev, PCI_BASE_ADDRESS_0 + i * 4, 0xffffffff);
		ops->read_dword(dev, PCI_BASE_ADDRESS_0 + i * 4, &bar);

		if (bar == 0) {
			continue;
		}
		if ((bar & 0xffff0000) == 0) {
			bar |= 0xffff0000;	/* for buggy device */
		}
		if (bar & PCI_BASE_ADDRESS_SPACE) {
			/* I/O space */
			*io_size += ~(bar & PCI_BASE_ADDRESS_IO_MASK) + 1;
		} else {
			/* memory space */
			if ((bar & PCI_BASE_ADDRESS_MEM_TYPE_MASK) ==
			    PCI_BASE_ADDRESS_MEM_TYPE_64) {
				is_64bits_memory = 1;
			}
			if (bar & PCI_BASE_ADDRESS_MEM_PREFETCH) {
				/* prefetchable memory space */
				*pmem_size +=
					~(bar & PCI_BASE_ADDRESS_MEM_MASK) + 1;
			} else {
				/* memory space */
				*mem_size +=
					~(bar & PCI_BASE_ADDRESS_MEM_MASK) + 1;
			}
		}
		if (is_64bits_memory) {
			i++;
		}
	}

	DPRFE;
	return;
}

static int
pciconf_alloc_bar(struct pci_dev *dev, struct pci_ops *ops, size_t num,
		  struct resource *io_resources[],
		  struct resource *mem_resources[],
		  struct resource *pmem_resources[])
{
	int res;
	int i;
	u32 bar;
	struct resource *new_res;
	u32 size;
	u16 command;
	int is_64bits_memory;
	DPRFS;

	for (i = 0; i < num; i++) {
		is_64bits_memory = 0;

		ops->write_dword(dev, PCI_BASE_ADDRESS_0 + i * 4, 0xffffffff);
		ops->read_dword(dev, PCI_BASE_ADDRESS_0 + i * 4, &bar);

		if (bar == 0) {
			continue;
		}
		if ((bar & 0xffff0000) == 0) {
			bar |= 0xffff0000;	/* for buggy device */
		}
		DPRFL(DPL_DEBUG, "bar[%02x:%02x.%x][%d] = 0x%08x\n",
		      dev->bus->number, PCI_SLOT(dev->devfn),
		      PCI_FUNC(dev->devfn), i, bar);
		if (bar & PCI_BASE_ADDRESS_SPACE) {
			/* I/O space */
			size = ~(bar & PCI_BASE_ADDRESS_IO_MASK) + 1;
			new_res = pciconf_malloc_res("PCI I/O resource",
						     IORESOURCE_IO);
			if (new_res == NULL) {
				res = -ENOMEM;
				goto err;
			}
			res = pciconf_allocate_resource(io_resources,
							new_res, size, size);
			if (res < 0) {
				goto err;
			}
			DPRFL(DPL_DEBUG,
			      "PCI[%02x:%02x.%x] allocated I/O resource "
			      "0x%08lx-0x%08lx\n",
			      dev->bus->number, PCI_SLOT(dev->devfn),
			      PCI_FUNC(dev->devfn),
			      new_res->start, new_res->end);
		} else {
			/* memory space */
			if ((bar & PCI_BASE_ADDRESS_MEM_TYPE_MASK) ==
			    PCI_BASE_ADDRESS_MEM_TYPE_64) {
				is_64bits_memory = 1;
			}
			size = ~(bar & PCI_BASE_ADDRESS_MEM_MASK) + 1;
			new_res = pciconf_malloc_res("PCI memory resource",
						     IORESOURCE_MEM);
			if (new_res == NULL) {
				res = -ENOMEM;
				goto err;
			}
			if (bar & PCI_BASE_ADDRESS_MEM_PREFETCH) {
				/* prefetchable memory space */
				new_res->flags |= IORESOURCE_PREFETCH;
				res = pciconf_allocate_resource(pmem_resources,
								new_res, size,
								size);
			} else {
				/* memory space */
				res = pciconf_allocate_resource(mem_resources,
								new_res, size,
								size);
			}				
			if (res < 0) {
				goto err;
			}
			DPRFL(DPL_DEBUG,
			      "PCI[%02x:%02x.%x] allocated memory resource "
			      "0x%08lx-0x%08lx\n",
			      dev->bus->number, PCI_SLOT(dev->devfn),
			      PCI_FUNC(dev->devfn),
			      new_res->start, new_res->end);
		}
		ops->write_dword(dev, PCI_BASE_ADDRESS_0 + i * 4,
				 new_res->start);
		if (is_64bits_memory) {
			i++;
			if (i < num) {
				ops->write_dword(dev,
						 PCI_BASE_ADDRESS_0 + i * 4,
						 0);
			}
		}
	}

	/* enable I/O and memory access */
	ops->read_word(dev, PCI_COMMAND, &command);
	command |= (PCI_COMMAND_IO | PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER);
	ops->write_word(dev, PCI_COMMAND, command);

	/* set latency timer */
	ops->write_byte(dev, PCI_LATENCY_TIMER, 0x80);

	DPRFE;
	return 0;

err:
	DPRFR;
	return res;
}

static void
pciconf_preset_bridge_win(struct pci_dev *dev, struct pci_ops *ops,
			  u32 io_size, u32 mem_size, u32 pmem_size)
{
	DPRFS;

	ops->write_byte(dev, PCI_IO_BASE, 0xff);
	ops->write_word(dev, PCI_IO_BASE_UPPER16, 0xffff & PCI_IO_RANGE_MASK);
	ops->write_byte(dev, PCI_IO_LIMIT,
			((io_size - 1) >> 8) & PCI_IO_RANGE_MASK);
	ops->write_word(dev, PCI_IO_LIMIT_UPPER16, (io_size - 1) >> 16);

	ops->write_word(dev, PCI_MEMORY_BASE, 0xffff & PCI_MEMORY_RANGE_MASK);
	ops->write_word(dev, PCI_MEMORY_LIMIT,
			((mem_size - 1) >> 16) & PCI_MEMORY_RANGE_MASK);
	ops->write_word(dev, PCI_PREF_MEMORY_BASE,
			0xffff & PCI_PREF_RANGE_MASK);
	ops->write_dword(dev, PCI_PREF_BASE_UPPER32, 0);
	ops->write_word(dev, PCI_PREF_MEMORY_LIMIT,
			((pmem_size - 1) >> 16) & PCI_PREF_RANGE_MASK);
	ops->write_dword(dev, PCI_PREF_LIMIT_UPPER32, 0);

	DPRFE;
	return;
}

static int
pciconf_alloc_bridge_win(struct pci_dev *dev, struct pci_ops *ops,
			 struct resource *io_resources[],
			 struct resource *mem_resources[],
			 struct resource *pmem_resources[])
{
	int res;
	u8 io_limit;
	u16 io_limit_upper16;
	u16 mem_limit;
	u16 pmem_limit;
	u32 io_size;
	u32 mem_size;
	u32 pmem_size;
	struct resource *new_io_resources[2];
	struct resource *new_mem_resources[2];
	struct resource *new_pmem_resources[2];
	u8 secondary_bus;
	DPRFS;

	/* allocate I/O window */
	ops->read_byte(dev, PCI_IO_LIMIT, &io_limit);
	ops->read_word(dev, PCI_IO_LIMIT_UPPER16, &io_limit_upper16);
	io_size = ((((io_limit & PCI_IO_RANGE_MASK) +
		     (~PCI_IO_RANGE_MASK + 1)) << 8) +
		   (io_limit_upper16 << 16));
	new_io_resources[0] =
		pciconf_malloc_res("PCI-to-PCI bridge I/O window",
				   IORESOURCE_IO);
	if (new_io_resources[0] == NULL) {
		res = -ENOMEM;
		goto err;
	}
	res = pciconf_allocate_resource(io_resources, new_io_resources[0],
					io_size,
					pciconf_align_pow2(io_size));
	if (res < 0) {
		goto err;
	}

	/* allocate memory window */
	ops->read_word(dev, PCI_MEMORY_LIMIT, &mem_limit);
	mem_size = (((mem_limit & PCI_MEMORY_RANGE_MASK) +
		     (~PCI_MEMORY_RANGE_MASK + 1)) << 16);
	new_mem_resources[0] =
		pciconf_malloc_res("PCI-to-PCI bridge memory window",
				   IORESOURCE_MEM);
	if (new_mem_resources[0] == NULL) {
		res = -ENOMEM;
		goto err;
	}
	res = pciconf_allocate_resource(mem_resources, new_mem_resources[0],
					mem_size,
					pciconf_align_pow2(mem_size));
	if (res < 0) {
		goto err;
	}

	/* allocate prefetchable memory window */
	ops->read_word(dev, PCI_PREF_MEMORY_LIMIT, &pmem_limit);
	pmem_size = (((pmem_limit & PCI_PREF_RANGE_MASK) +
		      (~PCI_PREF_RANGE_MASK + 1)) << 16);
	new_pmem_resources[0] =
		pciconf_malloc_res("PCI-to-PCI bridge prefetchable "
				   "memory window",
				   IORESOURCE_MEM | IORESOURCE_PREFETCH);
	if (new_pmem_resources[0] == NULL) {
		res = -ENOMEM;
		goto err;
	}
	res = pciconf_allocate_resource(pmem_resources, new_pmem_resources[0],
					pmem_size,
					pciconf_align_pow2(pmem_size));
	if (res < 0) {
		goto err;
	}

	/* set window registers */
	ops->write_byte(dev, PCI_IO_BASE,
			(new_io_resources[0]->start >> 8) & PCI_IO_RANGE_MASK);
	ops->write_word(dev, PCI_IO_BASE_UPPER16,
			(new_io_resources[0]->start >> 16));
	ops->write_byte(dev, PCI_IO_LIMIT,
			(((new_io_resources[0]->end - 1) >> 8) &
			 PCI_IO_RANGE_MASK));
	ops->write_word(dev, PCI_IO_LIMIT_UPPER16,
			(new_io_resources[0]->end - 1) >> 16);
	ops->write_word(dev, PCI_MEMORY_BASE,
			((new_mem_resources[0]->start >> 16) &
			 PCI_MEMORY_RANGE_MASK));
	ops->write_word(dev, PCI_MEMORY_LIMIT,
			(((new_mem_resources[0]->end - 1) >> 16) &
			 PCI_MEMORY_RANGE_MASK));
	ops->write_word(dev, PCI_PREF_MEMORY_BASE,
			((new_pmem_resources[0]->start >> 16) &
			 PCI_PREF_RANGE_MASK));
	ops->write_dword(dev, PCI_PREF_BASE_UPPER32, 0);
	ops->write_word(dev, PCI_PREF_MEMORY_LIMIT,
			(((new_pmem_resources[0]->end - 1) >> 16) &
			 PCI_PREF_RANGE_MASK));
	ops->write_dword(dev, PCI_PREF_LIMIT_UPPER32, 0);

	/* allocate resouces on sub bus */
	new_io_resources[1] = NULL;
	new_mem_resources[1] = NULL;
	new_pmem_resources[1] = NULL;
	ops->read_byte(dev, PCI_SECONDARY_BUS, &secondary_bus);
	res = pciconf_alloc_tree(secondary_bus, dev->bus, ops,
				 new_io_resources, new_mem_resources,
				 new_pmem_resources);
	if (res < 0) {
		goto err;
	}

	DPRFE;
	return 0;

err:
	DPRFR;
	return res;
}

static void
pciconf_preset_cb_win(struct pci_dev *dev, struct pci_ops *ops,
		      u32 io_size, u32 mem_size, u32 pmem_size)
{
	DPRFS;

	/**** TODO: use *_size */

	ops->write_dword(dev, PCI_CB_MEMORY_BASE_0, 0xffffffff);
	ops->write_dword(dev, PCI_CB_MEMORY_LIMIT_0, 0);
	ops->write_dword(dev, PCI_CB_MEMORY_BASE_1, 0xffffffff);
	ops->write_dword(dev, PCI_CB_MEMORY_LIMIT_1, 0);
	ops->write_dword(dev, PCI_CB_IO_BASE_0, 0xffffffff);
	ops->write_dword(dev, PCI_CB_IO_LIMIT_0, 0);
	ops->write_dword(dev, PCI_CB_IO_BASE_1, 0xffffffff);
	ops->write_dword(dev, PCI_CB_IO_LIMIT_1, 0);

	DPRFE;
	return;
}
